from __future__ import print_function

from os import path
from random import randint
import time

from twisted.internet import reactor
from twisted.internet.task import LoopingCall

# Check if we are running from the root directory
# If not, modify our path so that we can import IPv8
try:
    import ipv8
    del ipv8
except ImportError:
    import sys
    sys.path.append(path.abspath(path.join(path.dirname(__file__), "..")))

from ipv8_service import _COMMUNITIES, IPv8
from ipv8.configuration import get_default_configuration
from ipv8.community import Community
from ipv8.keyvault.crypto import ECCrypto
from ipv8.peer import Peer


START_TIME = time.time()
LOW_EDGE = 0
LOW_EDGE_PEER = None
INSTANCES = []


class MyCommunity(Community):
    master_peer = Peer(ECCrypto().generate_key(u"medium"))

    def started(self):
        def check_peers():
            global INSTANCES, LOW_EDGE, LOW_EDGE_PEER, START_TIME
            if self.get_peers():
                if LOW_EDGE and self.my_peer != LOW_EDGE_PEER:
                    for instance in INSTANCES:
                        instance.stop(False)
                    reactor.callFromThread(reactor.stop)
                    print("%.4f,%.4f" % (LOW_EDGE, time.time() - START_TIME))
                else:
                    LOW_EDGE = time.time() - START_TIME
                    LOW_EDGE_PEER = self.my_peer
        self.register_task("check_peers", LoopingCall(check_peers)).start(0.1, True)

    def create_introduction_response(self, lan_socket_address, socket_address, identifier, introduction=None):
        return super(MyCommunity, self).create_introduction_response(lan_socket_address, socket_address,
                                                                     identifier, introduction, b'1')

    def create_introduction_request(self, socket_address):
        return super(MyCommunity, self).create_introduction_request(socket_address, b'2')


_COMMUNITIES['MyCommunity'] = MyCommunity


for i in [1, 2]:
    configuration = get_default_configuration()
    configuration['keys'] = [{
        'alias': "my peer",
        'generation': u"medium",
        'file': u"ec%d.pem" % i
    }]
    configuration['port'] = 12000 + randint(0, 10000)
    configuration['overlays'] = [{
        'class': 'MyCommunity',
        'key': "my peer",
        'walkers': [{
            'strategy': "RandomWalk",
            'peers': 10,
            'init': {
                'timeout': 3.0
            }
        }],
        'initialize': {},
        'on_start': [('started', )]
    }]
    INSTANCES.append(IPv8(configuration))

reactor.run()



from __future__ import absolute_import

import logging
from random import random
from threading import Lock

from six import integer_types, text_type
from six.moves import xrange

from .taskmanager import TaskManager


class NumberCache(object):

    def __init__(self, request_cache, prefix, number):
        assert isinstance(number, integer_types), type(number)
        assert isinstance(prefix, text_type), type(prefix)

        super(NumberCache, self).__init__()
        self._logger = logging.getLogger(self.__class__.__name__)

        if request_cache.has(prefix, number):
            raise RuntimeError("This number is already in use '%s'" % number)

        self._prefix = prefix
        self._number = number

    @property
    def prefix(self):
        return self._prefix

    @property
    def number(self):
        return self._number

    @property
    def timeout_delay(self):
        return 10.0

    def on_timeout(self):
        raise NotImplementedError()

    def __str__(self):
        return "<%s %s-%d>" % (self.__class__.__name__, self.prefix, self.number)


class RandomNumberCache(NumberCache):

    def __init__(self, request_cache, prefix):
        assert isinstance(prefix, text_type), type(prefix)

        # find an unclaimed identifier
        number = RandomNumberCache.find_unclaimed_identifier(request_cache, prefix)
        super(RandomNumberCache, self).__init__(request_cache, prefix, number)

    @classmethod
    def find_unclaimed_identifier(cls, request_cache, prefix):
        for _ in xrange(1000):
            number = int(random() * 2 ** 16)
            if not request_cache.has(prefix, number):
                break
        else:
            raise RuntimeError("Could not find a number that isn't in use")

        return number


class RequestCache(TaskManager):

    def __init__(self):
        """
        Creates a new RequestCache instance.
        """
        super(RequestCache, self).__init__()

        self._logger = logging.getLogger(self.__class__.__name__)

        self._identifiers = dict()
        self.lock = Lock()
        self._shutdown = False

    def add(self, cache):
        """
        Add CACHE into this RequestCache instance.

        Returns CACHE when CACHE.identifier was not yet added, otherwise returns None.
        """
        assert isinstance(cache, NumberCache), type(cache)
        assert isinstance(cache.number, integer_types), type(cache.number)
        assert isinstance(cache.prefix, text_type), type(cache.prefix)
        assert isinstance(cache.timeout_delay, float), type(cache.timeout_delay)
        assert cache.timeout_delay > 0.0, cache.timeout_delay

        with self.lock:
            if self._shutdown:
                self._logger.warning("Dropping %s due to shutdown!", str(cache))
                return None

            identifier = self._create_identifier(cache.number, cache.prefix)
            if identifier in self._identifiers:
                self._logger.error("add with duplicate identifier \"%s\"", identifier)
                return None

            else:
                self._logger.debug("add %s", cache)
                self._identifiers[identifier] = cache
                self.register_task(cache, self._reactor.callLater(cache.timeout_delay, self._on_timeout, cache))
                return cache

    def has(self, prefix, number):
        """
        Returns True when IDENTIFIER is part of this RequestCache.
        """
        assert isinstance(number, integer_types), type(number)
        assert isinstance(prefix, text_type), type(prefix)
        return self._create_identifier(number, prefix) in self._identifiers

    def get(self, prefix, number):
        """
        Returns the Cache associated with IDENTIFIER when it exists, otherwise returns None.
        """
        assert isinstance(number, integer_types), type(number)
        assert isinstance(prefix, text_type), type(prefix)
        return self._identifiers.get(self._create_identifier(number, prefix))

    def pop(self, prefix, number):
        """
        Returns the Cache associated with IDENTIFIER, and removes it from this RequestCache, when it exists, otherwise
        raises a KeyError exception.
        """
        assert isinstance(number, integer_types), type(number)
        assert isinstance(prefix, text_type), type(prefix)

        identifier = self._create_identifier(number, prefix)
        cache = self._identifiers.pop(identifier)
        self.cancel_pending_task(cache)
        return cache

    def _on_timeout(self, cache):
        """
        Called CACHE.timeout_delay seconds after CACHE was added to this RequestCache.

        _on_timeout is called for every Cache, except when it has been popped before the timeout expires.  When called
        _on_timeout will CACHE.on_timeout().
        """
        assert isinstance(cache, NumberCache), type(cache)

        self._logger.debug("timeout on %s", cache)

        # the on_timeout call could have already removed the identifier from the cache using pop
        identifier = self._create_identifier(cache.number, cache.prefix)
        if identifier in self._identifiers:
            del self._identifiers[identifier]

        cache.on_timeout()

        self.cancel_pending_task(cache)

    def _create_identifier(self, number, prefix):
        return u"%s:%d" % (prefix, number)

    def clear(self):
        """
        Clear the cache, canceling all pending tasks.

        """
        self._logger.debug("Clearing %s [%s]", self, len(self._identifiers))
        self.cancel_all_pending_tasks()
        self._identifiers.clear()

    def shutdown(self):
        """
        Clear the cache, cancel all pending tasks and disallow new caches being added.
        """
        with self.lock:
            self._shutdown = True
            self.clear()



from __future__ import absolute_import

from base64 import b64encode

from .base_endpoint import BaseEndpoint


class NetworkEndpoint(BaseEndpoint):
    """
    This endpoint is responsible for handing all requests regarding the state of the network.
    """

    def __init__(self, session):
        super(NetworkEndpoint, self).__init__()
        self.session = session

    def retrieve_peers(self):
        network = self.session.network
        peer_list = network.verified_peers[:]
        return {
            b64encode(peer.mid): {
                "ip": peer.address[0],
                "port": peer.address[1],
                "public_key": b64encode(peer.public_key.key_to_bin()),
                "services": [b64encode(s) for s in network.get_services_for_peer(peer)]
            }
            for peer in peer_list
        }

    def render_GET(self, request):
        return self.twisted_dumps({"peers": self.retrieve_peers()})



from __future__ import absolute_import

from binascii import unhexlify

from twisted.trial import unittest

from .....attestation.wallet.primitives.boneh import (FP2Value, bilinear_group, decode, encode, generate_keypair,
                                                      generate_prime, get_good_wp, is_good_wp)
from .....attestation.wallet.primitives.structs import BonehPrivateKey


class TestBoneh(unittest.TestCase):

    private_key = BonehPrivateKey.unserialize(unhexlify("0142018eceb3e03820006219a5c5a959abfd40b042c381ff894f7c3d62"
                                                        "5ee9a02302dcdeda322aa8372b66c8d9b4df981e96eb1e4f7dacda5bca"
                                                        "7399d864941a4b3ab55101415e009ff86a3288953321922929aaaaecdf"
                                                        "ba4663004f3b6caa1a32df6af462659a7bb4659f2325eba7e8518b526b"
                                                        "e30b0002d304429b21643154f8cec968d6156e0141608dd4640f02a8dd"
                                                        "1b98c012df578840cabadbe985e50ba5481e4f5ea9640fb363c4772998"
                                                        "5405ee29cacc9786ba747d9a26a9e98752d728c459bd3b8214f302be01"
                                                        "4143286cfe8bf4b0b0283714068ced3bb6be72e5a913ecfe5827cbee71"
                                                        "ce11a69f6a0ead8270430c3717606c1e93b7e60d2e426f26bf1ad9e81f"
                                                        "d260da09e47281fb0142017cdb49d11264889ce2d989f540196f3bfb01"
                                                        "02021d0426667e137c76b5c993406681f6b2335b2734ee6c6e251706f6"
                                                        "c2bf0c4fe53d0956410512f90465a3f777810140b6f72590286e1bb3e3"
                                                        "64d3cd3c6898e3df165f3361f76cf42e8361ae2a4547f73947a8e871e0"
                                                        "6bff71401841e0db916571eb947f03d5b236dee592286b7d5a3f0120cd"
                                                        "8489a2b149edce500b4357aa9ed0b6c220d7a7c16706531671e30d0334"
                                                        "2ef3"))

    def test_generate_prime(self):
        """
        Check if the next prime (= l * n - 1 = 2 mod 3) after 10 is 29.
        """
        self.assertEqual(generate_prime(10), 29)

    def test_bilinear_group(self):
        """
        Check if a bilinear group can be created.
        """
        self.assertEqual(bilinear_group(10, 29, 4, 5, 4, 5), FP2Value(29, 19, 5))

    def test_bilinear_group_torsion_point(self):
        """
        Check if a bilinear group returns 0 if there is no possible pairing.
        """
        self.assertEqual(bilinear_group(10, 29, 2, 3, 2, 3), FP2Value(29))

    def test_is_good_weil_pairing(self):
        """
        Check if is_good_wp returns True for 26 + 17x with n = 10
        """
        self.assertTrue(is_good_wp(10, FP2Value(29, 26, 17)))

    def test_is_bad_weil_pairing(self):
        """
        Check if is_good_wp returns False for 0, 1 and x with n = 10
        """
        self.assertFalse(is_good_wp(10, FP2Value(29)))
        self.assertFalse(is_good_wp(10, FP2Value(29, 1)))
        self.assertFalse(is_good_wp(10, FP2Value(29, b=1)))

    def test_get_good_weil_pairing(self):
        """
        Check if get_good_wp returns a proper Weil pairing for n = 10, p = 29.
        """
        _, wp = get_good_wp(10, 29)

        self.assertTrue(is_good_wp(10, wp))

    def test_encoding_random(self):
        """
        Check if the same value is encoded with a random masking.
        """
        pk = TestBoneh.private_key.public_key()

        self.assertNotEqual(encode(pk, 0), encode(pk, 0))

    def test_decoding_same(self):
        """
        Check if values are encoded with a random masking.
        """
        pk = TestBoneh.private_key.public_key()
        a = encode(pk, 0)
        b = encode(pk, 0)

        self.assertEqual(decode(TestBoneh.private_key, [0], a), decode(TestBoneh.private_key, [0], b))

    def test_decoding_after_homomorphic_add(self):
        """
        Check if values can still be decrypted after a homomorphic add.
        """
        pk = TestBoneh.private_key.public_key()
        a = encode(pk, 1)
        b = encode(pk, 3)

        self.assertEqual(decode(TestBoneh.private_key, [4], a * b), 4)

    def test_decoding_out_of_space(self):
        """
        Check if decode return None if the message is outside of the allowed space.
        """
        pk = TestBoneh.private_key.public_key()

        self.assertIsNone(decode(TestBoneh.private_key, [0], encode(pk, 1)))

    def test_generate_keypair(self):
        """
        Check if we can create a new keypair.
        """
        pk, sk = generate_keypair(32)

        self.assertEqual(pk.p, sk.p)
        self.assertEqual(pk.g, sk.g)
        self.assertEqual(pk.h, sk.h)
        self.assertEqual(decode(sk, [0, 1, 2], encode(pk, 0)), 0)
        self.assertEqual(decode(sk, [0, 1, 2], encode(pk, 1)), 1)
        self.assertEqual(decode(sk, [0, 1, 2], encode(pk, 2)), 2)



from __future__ import absolute_import

from hashlib import sha256, sha512


def sha512_as_int(value):
    """
    Convert a SHA512 hash to an integer.
    """
    out = 0
    hashed = sha512(value).digest()
    for i in range(len(hashed)):
        out <<= 8
        out |= ord(hashed[i:i + 1])
    return out


def sha256_as_int(value):
    """
    Convert a SHA256 hash to an integer.
    """
    out = 0
    hashed = sha256(value).digest()
    for i in range(len(hashed)):
        out <<= 8
        out |= ord(hashed[i:i + 1])
    return out


def sha256_4_as_int(value):
    """
    Convert a SHA256 4 byte hash to an integer.
    """
    out = 0
    hashed = sha256(value).digest()[:4]
    for i in range(len(hashed)):
        out <<= 8
        out |= ord(hashed[i:i + 1])
    return out



from __future__ import absolute_import, division

from binascii import hexlify, unhexlify
from struct import pack, unpack

from six.moves import xrange

from ..pengbaorange.boudot import EL, SQR
from ..primitives.boneh import decode, encode
from ..primitives.structs import BonehPublicKey, ipack, iunpack
from ..primitives.value import FP2Value
from ...identity_formats import Attestation


def _serialize_fp2value(value):
    normalized = value.wp_compress()
    return ipack(normalized.a) + ipack(normalized.b)


def _unserialize_fp2value(mod, value):
    rem = value
    a, rem = iunpack(rem)
    b, rem = iunpack(rem)
    return FP2Value(mod, a, b), rem


class PengBaoCommitment(object):

    def __init__(self, c, c1, c2, ca, ca1, ca2, ca3, caa):  # pylint: disable=R0913
        self.c = c
        self.c1 = c1
        self.c2 = c2
        self.ca = ca
        self.ca1 = ca1
        self.ca2 = ca2
        self.ca3 = ca3
        self.caa = caa

    def serialize(self):
        return (ipack(self.c.mod) + _serialize_fp2value(self.c) + _serialize_fp2value(self.c1)
                + _serialize_fp2value(self.c2) + _serialize_fp2value(self.ca) + _serialize_fp2value(self.ca1)
                + _serialize_fp2value(self.ca2) + _serialize_fp2value(self.ca3) + _serialize_fp2value(self.caa))

    @classmethod
    def unserialize(cls, s):
        """
        :param s: the string to unserialize
        :type s: str
        :return: the unserialized commitment
        :rtype: PengBaoCommitment
        """
        mod, rem = iunpack(s)
        c, rem = _unserialize_fp2value(mod, rem)
        c1, rem = _unserialize_fp2value(mod, rem)
        c2, rem = _unserialize_fp2value(mod, rem)
        ca, rem = _unserialize_fp2value(mod, rem)
        ca1, rem = _unserialize_fp2value(mod, rem)
        ca2, rem = _unserialize_fp2value(mod, rem)
        ca3, rem = _unserialize_fp2value(mod, rem)
        caa, rem = _unserialize_fp2value(mod, rem)
        return cls(c, c1, c2, ca, ca1, ca2, ca3, caa), rem


class PengBaoCommitmentPrivate(object):

    MSGSPACE = list(range(256))

    def __init__(self, m1, m2, m3, r1, r2, r3):
        self.m1 = m1
        self.m2 = m2
        self.m3 = m3
        self.r1 = r1
        self.r2 = r2
        self.r3 = r3

    def generate_response(self, s, t):
        """
        Given s and t and our private information, generate the appropriate response.

        The calling function should make sure s and t are large integers (at least larger than 0).

        :param s: the challenge value s
        :type s: int
        :param t: the challenge value t
        :type t: int
        :return: the challenge response
        :rtype: int
        """
        return (s * self.m1 + self.m2 + self.m3,
                self.m1 + t * self.m2 + self.m3,
                s * self.r1 + self.r2 + self.r3,
                self.r1 + t * self.r2 + self.r3)

    def serialize(self):
        return ipack(self.m1) + ipack(self.m2) + ipack(self.m3) + ipack(self.r1) + ipack(self.r2) + ipack(self.r3)

    @classmethod
    def unserialize(cls, s):
        """
        :param s: the string to unserialize
        :type s: str
        :return: the unserialized private data
        :rtype: PengBaoCommitmentPrivate
        """
        m1, rem = iunpack(s)
        m2, rem = iunpack(rem)
        m3, rem = iunpack(rem)
        r1, rem = iunpack(rem)
        r2, rem = iunpack(rem)
        r3, rem = iunpack(rem)
        return cls(m1, m2, m3, r1, r2, r3), rem

    def encode(self, PK):
        serialized = self.serialize()
        hex_serialized = hexlify(serialized)
        serialized_encodings = pack(">B", len(hex_serialized) // 2)
        for i in xrange(0, len(hex_serialized), 2):
            intval = int(hex_serialized[i:i + 2], 16)
            serialized_encodings += _serialize_fp2value(encode(PK, intval))
        return serialized_encodings

    @classmethod
    def decode(cls, SK, s):
        serialized = b""
        count, = unpack(">B", s[0:1])
        rem = s[1:]
        for _ in xrange(count):
            unpacked, rem = _unserialize_fp2value(SK.g.mod, rem)
            hexed = hex(decode(SK, cls.MSGSPACE, unpacked))[2:]
            if hexed.endswith('L'):
                hexed = hexed[:-1]
            if len(hexed) % 2 == 1:
                hexed = '0' + hexed
            serialized += unhexlify(hexed)
        return cls.unserialize(serialized)[0]


class PengBaoPublicData(object):
    """
    Public data required to verify a Peng Bao proof.
    """

    def __init__(self, PK, bitspace, commitment, el, sqr1, sqr2):
        """
        :param PK: the BonehPublicKey of the owner
        :type PK: BonehPublicKey
        :param bitspace: the bitspace for the commitment message
        :type bitspace: int
        :param commitment: the range commitment
        :type commitment: PengBaoCommitment
        :param el: the Boudot equality commitment
        :type el: EL
        :param sqr1: the first Boudot square commitment
        :type sqr1: SQR
        :param sqr2: the second Boudot square commitment
        :type sqr2: SQR
        """
        self.PK = PK
        self.bitspace = bitspace
        self.commitment = commitment
        self.el = el
        self.sqr1 = sqr1
        self.sqr2 = sqr2

    def check(self, a, b, s, t, x, y, u, v):  # pylint: disable=R0913
        out = True
        out &= self.el.check(self.PK.g, self.PK.h, self.commitment.c1, self.PK.h, self.commitment.c2,
                             self.commitment.ca)
        out &= self.sqr1.check(self.commitment.ca, self.PK.h, self.commitment.caa)
        out &= self.sqr2.check(self.PK.g, self.PK.h, self.commitment.ca3)
        out &= self.commitment.c1 == self.commitment.c // self.PK.g.intpow(a - 1)
        out &= self.commitment.c2 == self.PK.g.intpow(b + 1) // self.commitment.c
        out &= self.commitment.caa == self.commitment.ca1 * self.commitment.ca2 * self.commitment.ca3
        out &= (self.PK.g.intpow(x) * self.PK.h.intpow(u)) == (self.commitment.ca1.intpow(s) * self.commitment.ca2
                                                               * self.commitment.ca3)
        out &= (self.PK.g.intpow(y) * self.PK.h.intpow(v)) == (self.commitment.ca1 * self.commitment.ca2.intpow(t)
                                                               * self.commitment.ca3)
        out &= x > 0
        out &= y > 0
        return out

    def serialize(self):
        """
        Serialize this Attestation to a string.

        :return: the serialized form of this attestation
        :rtype: str
        """
        return (self.PK.serialize() + ipack(self.bitspace) + self.commitment.serialize() + self.el.serialize()
                + self.sqr1.serialize() + self.sqr2.serialize())

    @classmethod
    def unserialize(cls, s):
        """
        Given a string, create an Attestation object.

        :param s: the string to unserialize
        :type s: str
        :return: the attestation object
        :rtype: PengBaoPublicData
        """
        rem = s
        pk = BonehPublicKey.unserialize(rem)
        rem = rem[len(pk.serialize()):]
        bitspace, rem = iunpack(rem)
        commitment, rem = PengBaoCommitment.unserialize(rem)
        el, rem = EL.unserialize(rem)
        sqr1, rem = SQR.unserialize(rem)
        sqr2, rem = SQR.unserialize(rem)
        return cls(pk, bitspace, commitment, el, sqr1, sqr2), rem


class PengBaoAttestation(Attestation):

    def __init__(self, publicdata, privatedata, id_format=None):
        """
        :type publicdata: PengBaoPublicData
        :type privatedata: PengBaoCommitmentPrivate or None
        """
        super(PengBaoAttestation, self).__init__()
        self.publicdata = publicdata
        self.privatedata = privatedata
        self.id_format = id_format
        self.PK = publicdata.PK

    def serialize(self):
        """
        Serialize this Attestation to a string.

        :return: the serialized form of this attestation
        :rtype: str
        """
        return self.publicdata.serialize()

    def serialize_private(self, PK):
        """
        Serialize this Attestation to a string, include shared secrets (not to be published!).

        :param PK: the public key to encode for
        :return: the serialized form of this attestation
        :rtype: str
        """
        assert self.privatedata
        return self.publicdata.serialize() + self.privatedata.encode(PK)

    @classmethod
    def unserialize(cls, s, id_format=None):
        """
        Given a string, create an Attestation object.

        :param s: the string to unserialize
        :type s: str
        :param id_format: the identity format
        :type id_format: str
        :return: the attestation object
        :rtype: Attestation
        """
        publicdata, _ = PengBaoPublicData.unserialize(s)
        return cls(publicdata, None, id_format)

    @classmethod
    def unserialize_private(cls, SK, s, id_format=None):
        """
        Given a string, create an Attestation object.
        The input contains shared secrets not to be published.

        :param SK: the secret key to decode with
        :param s: the string to unserialize
        :type s: str
        :param id_format: the identity format
        :type id_format: str
        :return: the attestation object
        :rtype: Attestation
        """
        publicdata, rem = PengBaoPublicData.unserialize(s)
        privatedata = PengBaoCommitmentPrivate.decode(SK, rem)

        return cls(publicdata, privatedata, id_format)



from __future__ import absolute_import

from base64 import b64encode
from struct import unpack
from time import time

from .keyvault.crypto import default_eccrypto
from .keyvault.keys import Key


class Peer(object):

    def __init__(self, key, address=("0.0.0.0", 0), intro=True):
        """
        Create a new Peer.

        :param key: the peer's Key (mostly public) or public key bin
        :param lan_address: the (IP, port) tuple of this peer on its LAN
        :param wan_address: the (IP, port) tuple of this peer on its WAN
        :param intro: is this peer suggested to us (otherwise it contacted us)
        """
        if not isinstance(key, Key):
            self.key = default_eccrypto.key_from_public_bin(key)
        else:
            self.key = key
        self.mid = self.key.key_to_hash()
        self.public_key = self.key.pub()
        self.address = address
        self.last_response = 0 if intro else time()
        self._lamport_timestamp = 0

    def update_clock(self, timestamp):
        """
        Update the Lamport timestamp for this peer. The Lamport clock dictates that the current timestamp is
        the maximum of the last known and the most recently delivered timestamp. This is useful when messages
        are delivered asynchronously.

        We also keep a real time timestamp of the last received message for timeout purposes.

        :param timestamp: a received timestamp
        """
        self._lamport_timestamp = max(self._lamport_timestamp, timestamp)
        self.last_response = time()  # This is in seconds since the epoch

    def get_lamport_timestamp(self):
        return self._lamport_timestamp

    def __hash__(self):
        as_long, = unpack(">Q", self.mid[:8])
        return as_long

    def __eq__(self, other):
        if not isinstance(other, Peer):
            return False
        return (self.public_key.key_to_bin() == other.public_key.key_to_bin())

    def __str__(self):
        return 'Peer<%s:%d, %s>' % (self.address[0], self.address[1], b64encode(self.mid))



from __future__ import absolute_import
from __future__ import division

from base64 import encodestring
from binascii import unhexlify

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.asymmetric.utils import decode_dss_signature

from .. import NEW_CRYPTOGRAPHY_SIGN_VERSION
from ...keyvault.public.m2crypto import M2CryptoPK
from ...keyvault.keys import PrivateKey


class M2CryptoSK(PrivateKey, M2CryptoPK):
    """
    A pyca implementation of a secret key, backwards compatible with Dispersy M2Crypto secret keys.
    """

    def __init__(self, curve=None, keystring=None, filename=None):
        """
        Create a new M2Crypto secret key. Optionally load it from a string representation (in a file)
        or generate it using some curve.

        :param curve: the EllipticCurve object specifying the curve
        :param keystring: the string to load the key from
        :param filename: the filename to load the key from
        """
        if curve:
            self.ec = ec.generate_private_key(curve, default_backend())

        elif keystring:
            self.ec = self.key_from_pem(b"-----BEGIN EC PRIVATE KEY-----\n%s-----END EC PRIVATE KEY-----\n" %
                                        encodestring(keystring))

        elif filename:
            with open(filename, 'rb') as keyfile:
                self.ec = self.key_from_pem(keyfile.read())

    def pub(self):
        """
        Get the public key for this secret key.
        """
        return M2CryptoPK(ec_pub=self.ec.public_key())

    def has_secret_key(self):
        """
        Is this is a secret key (yes it is)?
        """
        return True

    def key_to_pem(self):
        "Convert a key to the PEM format."

        return self.ec.private_bytes(serialization.Encoding.PEM,
                                     serialization.PrivateFormat.TraditionalOpenSSL,
                                     serialization.NoEncryption())

    def key_from_pem(self, pem):
        """
        Load a key from a pem certificate.

        :param pem: the PEM formatted private key
        """
        return serialization.load_pem_private_key(pem, password=None, backend=default_backend())

    def signature(self, msg):
        """
        Create a signature for a message in a backwards compatible fashion

        :param msg: the message to sign
        """
        # Create the pyca signature
        if NEW_CRYPTOGRAPHY_SIGN_VERSION:
            signature = self.ec.sign(msg, ec.ECDSA(hashes.SHA1()))
        else:
            signer = self.ec.signer(ec.ECDSA(hashes.SHA1()))
            signer.update(msg)
            signature = signer.finalize()
        # Decode the DSS r and s variables from the pyca signature
        # We are going to turn these longs into (binary) string format
        r, s = decode_dss_signature(signature)
        # Convert the r and s to a valid hex representation
        r = hex(r).rstrip("L").lstrip("0x") or "0"
        s = hex(s).rstrip("L").lstrip("0x") or "0"
        # We want bytes: one byte is two nibbles:
        # Prefix with a 0 if the result is of odd length
        if len(r) % 2 == 1:
            r = "0" + r
        if len(s) % 2 == 1:
            s = "0" + s
        # Now we can turn this into a binary string
        r = unhexlify(r)
        s = unhexlify(s)
        key_len = self.get_signature_length() // 2
        # For easy decoding, prepend 0 to r and s until they are of >equal length<
        return b"".join((b"\x00" * (key_len - len(r)), r, b"\x00" * (key_len - len(s)), s))



from __future__ import absolute_import

from ...attestation.trustchain.community import TrustChainCommunity
from ...dht.discovery import DHTDiscoveryCommunity
from ...keyvault.crypto import default_eccrypto
from ...messaging.interfaces.statistics_endpoint import StatisticsEndpoint
from ...peer import Peer
from ...peerdiscovery.network import Network
from .endpoint import AutoMockEndpoint
from .discovery import MockWalk


class MockIPv8(object):

    def __init__(self, crypto_curve, overlay_class, create_trustchain=False, create_dht=False, enable_statistics=False,
                 *args, **kwargs):
        self.endpoint = AutoMockEndpoint()
        self.endpoint.open()

        # Enable statistics
        if enable_statistics:
            self.endpoint = StatisticsEndpoint(self, self.endpoint)

        self.network = Network()
        self.my_peer = Peer(default_eccrypto.generate_key(crypto_curve), self.endpoint.wan_address)

        # Load a TrustChain community if specified
        self.trustchain = None
        if create_trustchain:
            self.trustchain = TrustChainCommunity(self.my_peer, self.endpoint, self.network,
                                                  working_directory=u":memory:")
            kwargs.update({'trustchain': self.trustchain})

        # Load a DHT community if specified
        self.dht = None
        if create_dht:
            self.dht = DHTDiscoveryCommunity(self.my_peer, self.endpoint, self.network)
            kwargs.update({'dht': self.dht})

        self.overlay = overlay_class(self.my_peer, self.endpoint, self.network, *args, **kwargs)
        self.overlay._use_main_thread = False
        self.discovery = MockWalk(self.overlay)

        self.overlay.my_estimated_wan = self.endpoint.wan_address
        self.overlay.my_estimated_lan = self.endpoint.lan_address

        if enable_statistics:
            self.endpoint.enable_community_statistics(self.overlay.get_prefix(), True)

    def unload(self):
        self.endpoint.close()
        self.overlay.unload()
        if self.trustchain:
            self.trustchain.unload()
        if self.dht:
            self.dht.unload()



from __future__ import absolute_import
from __future__ import division

import logging
import sys
import time
from base64 import b64decode
from os.path import isfile
from threading import RLock
from traceback import format_exception

from twisted.internet import reactor
from twisted.internet.defer import DeferredList, inlineCallbacks, maybeDeferred
from twisted.internet.task import LoopingCall

if hasattr(sys.modules['__main__'], "IPv8"):
    sys.modules[__name__] = sys.modules['__main__']
else:
    if __name__ == '__main__' or __name__ == 'ipv8_service':
        from ipv8.messaging.interfaces.statistics_endpoint import StatisticsEndpoint
        from ipv8.attestation.identity.community import IdentityCommunity
        from ipv8.attestation.trustchain.community import TrustChainCommunity, TrustChainTestnetCommunity
        from ipv8.attestation.wallet.community import AttestationCommunity
        from ipv8.keyvault.crypto import default_eccrypto
        from ipv8.keyvault.private.m2crypto import M2CryptoSK
        from ipv8.messaging.anonymization.community import TunnelCommunity
        from ipv8.messaging.anonymization.endpoint import TunnelEndpoint
        from ipv8.messaging.anonymization.hidden_services import HiddenTunnelCommunity
        from ipv8.messaging.interfaces.udp.endpoint import UDPEndpoint
        from ipv8.peer import Peer
        from ipv8.peerdiscovery.community import DiscoveryCommunity
        from ipv8.peerdiscovery.discovery import EdgeWalk, RandomWalk
        from ipv8.peerdiscovery.network import Network
        from ipv8.dht.discovery import DHTDiscoveryCommunity
    else:
        from .ipv8.messaging.interfaces.statistics_endpoint import StatisticsEndpoint
        from .ipv8.attestation.identity.community import IdentityCommunity
        from .ipv8.attestation.trustchain.community import TrustChainCommunity, TrustChainTestnetCommunity
        from .ipv8.attestation.wallet.community import AttestationCommunity
        from .ipv8.keyvault.crypto import default_eccrypto
        from .ipv8.keyvault.private.m2crypto import M2CryptoSK
        from .ipv8.messaging.anonymization.community import TunnelCommunity
        from .ipv8.messaging.anonymization.endpoint import TunnelEndpoint
        from .ipv8.messaging.anonymization.hidden_services import HiddenTunnelCommunity
        from .ipv8.messaging.interfaces.udp.endpoint import UDPEndpoint
        from .ipv8.peer import Peer
        from .ipv8.peerdiscovery.community import DiscoveryCommunity
        from .ipv8.peerdiscovery.discovery import EdgeWalk, RandomWalk
        from .ipv8.peerdiscovery.network import Network
        from .ipv8.dht.discovery import DHTDiscoveryCommunity

    _COMMUNITIES = {
        'AttestationCommunity': AttestationCommunity,
        'DiscoveryCommunity': DiscoveryCommunity,
        'HiddenTunnelCommunity': HiddenTunnelCommunity,
        'IdentityCommunity': IdentityCommunity,
        'TrustChainCommunity': TrustChainCommunity,
        'TunnelCommunity': TunnelCommunity,
        'DHTDiscoveryCommunity': DHTDiscoveryCommunity,
        'TrustChainTestnetCommunity': TrustChainTestnetCommunity,
    }

    _WALKERS = {
        'EdgeWalk': EdgeWalk,
        'RandomWalk': RandomWalk
    }

    class IPv8(object):

        def __init__(self, configuration, endpoint_override=None, enable_statistics=False, extra_communities=None):
            if endpoint_override:
                self.endpoint = endpoint_override
            else:
                self.endpoint = UDPEndpoint(port=configuration['port'], ip=configuration['address'])
                self.endpoint.open()
                if enable_statistics:
                    self.endpoint = StatisticsEndpoint(self, self.endpoint)
                if any([overlay.get('initialize', {}).get('anonymize') for overlay in configuration['overlays']]):
                    self.endpoint = TunnelEndpoint(self.endpoint)

            self.network = Network()

            # Load/generate keys
            self.keys = {}
            for key_block in configuration['keys']:
                if key_block['file'] and isfile(key_block['file']):
                    with open(key_block['file'], 'rb') as f:
                        content = f.read()
                        try:
                            # IPv8 Standardized bin format
                            self.keys[key_block['alias']] = Peer(default_eccrypto.key_from_private_bin(content))
                        except ValueError:
                            try:
                                # Try old Tribler M2Crypto PEM format
                                content = b64decode(content[31:-30].replace('\n', ''))
                                peer = Peer(M2CryptoSK(keystring=content))
                                peer.mid  # This will error out if the keystring is not M2Crypto
                                self.keys[key_block['alias']] = peer
                            except:
                                # Try old LibNacl format
                                content = "LibNaCLSK:" + content
                                self.keys[key_block['alias']] = Peer(default_eccrypto.key_from_private_bin(content))
                else:
                    self.keys[key_block['alias']] = Peer(default_eccrypto.generate_key(key_block['generation']))
                    if key_block['file']:
                        with open(key_block['file'], 'wb') as f:
                            f.write(self.keys[key_block['alias']].key.key_to_bin())

            # Setup logging
            logging.basicConfig(**configuration['logger'])

            self.overlay_lock = RLock()
            self.strategies = []
            self.overlays = []

            for overlay in configuration['overlays']:
                overlay_class = _COMMUNITIES.get(overlay['class'], (extra_communities or {}).get(overlay['class']))
                my_peer = self.keys[overlay['key']]
                overlay_instance = overlay_class(my_peer, self.endpoint, self.network, **overlay['initialize'])
                self.overlays.append(overlay_instance)
                for walker in overlay['walkers']:
                    strategy_class = _WALKERS.get(walker['strategy'],
                                                  overlay_instance.get_available_strategies().get(walker['strategy']))
                    args = walker['init']
                    target_peers = walker['peers']
                    self.strategies.append((strategy_class(overlay_instance, **args), target_peers))
                for config in overlay['on_start']:
                    reactor.callWhenRunning(getattr(overlay_instance, config[0]), *config[1:])

            self.state_machine_lc = LoopingCall(self.on_tick)
            self.state_machine_lc.start(configuration['walker_interval'], False)

        def on_tick(self):
            if self.endpoint.is_open():
                with self.overlay_lock:
                    smooth = self.state_machine_lc.interval // len(self.strategies) if self.strategies else 0
                    ticker = len(self.strategies)
                    for strategy, target_peers in self.strategies:
                        peer_count = len(strategy.overlay.get_peers())
                        start_time = time.time()
                        if (target_peers == -1) or (peer_count < target_peers):
                            # We wrap the take_step into a general except as it is prone to programmer error.
                            try:
                                strategy.take_step()
                            except:
                                logging.error("Exception occurred while trying to walk!\n"
                                              + ''.join(format_exception(*sys.exc_info())))
                        ticker -= 1 if ticker else 0
                        sleep_time = smooth - (time.time() - start_time)
                        if ticker and sleep_time > 0.01:
                            time.sleep(sleep_time)

        def unload_overlay(self, instance):
            with self.overlay_lock:
                self.overlays = [overlay for overlay in self.overlays if overlay != instance]
                self.strategies = [(strategy, target_peers) for (strategy, target_peers) in self.strategies
                                   if strategy.overlay != instance]
                return maybeDeferred(instance.unload)

        @inlineCallbacks
        def stop(self, stop_reactor=True):
            self.state_machine_lc.stop()
            with self.overlay_lock:
                unload_list = [self.unload_overlay(overlay) for overlay in self.overlays[:]]
                yield DeferredList(unload_list)
                yield self.endpoint.close()
            if stop_reactor:
                reactor.callFromThread(reactor.stop)


if __name__ == '__main__':
    from twisted.plugins.ipv8_plugin import Options, service_maker

    options = Options()
    Options.parseOptions(options, sys.argv[1:])
    service_maker.makeService(options)
    reactor.run()



from __future__ import absolute_import

from ..base import TestBase
from ...dht.trie import Trie


class TestNode(TestBase):

    def setUp(self):
        super(TestNode, self).setUp()
        self.trie = Trie(u'01')

    def test_get_set_del(self):
        self.trie[u'0'] = 1
        self.assertEqual(self.trie[u'0'], 1)

        self.trie[u'01'] = 2
        self.assertEqual(self.trie[u'01'], 2)

        self.trie[u'111'] = 4
        self.assertEqual(self.trie[u'111'], 4)

        self.trie[u'111'] = 3
        self.assertEqual(self.trie[u'111'], 3)

        with self.assertRaises(KeyError):
            # pylint: disable=W0104
            self.trie[u'000']

        del self.trie[u'111']
        with self.assertRaises(KeyError):
            # pylint: disable=W0104
            self.trie[u'111']

        with self.assertRaises(KeyError):
            del self.trie[u'1111']

    def test_longest_prefix(self):
        with self.assertRaises(KeyError):
            self.trie.longest_prefix(u'111')
        self.assertEqual(self.trie.longest_prefix(u'111', default=None), None)

        self.trie[u'0'] = 1
        self.trie[u'01'] = 2
        self.trie[u'111'] = 3

        self.assertEqual(self.trie.longest_prefix(u'000'), u'0')
        self.assertEqual(self.trie.longest_prefix(u'111'), u'111')

    def test_longest_prefix_item(self):
        with self.assertRaises(KeyError):
            self.trie.longest_prefix_item(u'111')
        self.assertEqual(self.trie.longest_prefix_item(u'111', default=None), None)

        self.trie[u'0'] = 1
        self.trie[u'01'] = 2
        self.trie[u'111'] = 3

        self.assertEqual(self.trie.longest_prefix_item(u'000'), (u'0', 1))
        self.assertEqual(self.trie.longest_prefix_item(u'111'), (u'111', 3))

    def test_longest_prefix_value(self):
        with self.assertRaises(KeyError):
            self.trie.longest_prefix_value(u'111')
        self.assertEqual(self.trie.longest_prefix_value(u'111', default=None), None)

        self.trie[u'0'] = 1
        self.trie[u'01'] = 2
        self.trie[u'111'] = 3

        self.assertEqual(self.trie.longest_prefix_value(u'000'), 1)
        self.assertEqual(self.trie.longest_prefix_value(u'111'), 3)

    def test_suffixes(self):
        self.assertEqual(self.trie.suffixes(u'111'), [])

        self.trie[u'0'] = 1
        self.trie[u'01'] = 2
        self.trie[u'111'] = 3
        self.trie[u'111111'] = 4

        self.assertEqual(self.trie.suffixes(u'0'), [u'', u'1'])
        self.assertEqual(self.trie.suffixes(u'11'), [u'1', u'1111'])

    def test_values(self):
        self.trie['0'] = 1
        self.trie['01'] = 2
        self.trie['111'] = 3

        self.assertSetEqual(set(self.trie.values()), set([1, 2, 3]))
        self.assertSetEqual(set(self.trie.values()), set([1, 2, 3]))



from __future__ import absolute_import

from collections import defaultdict

from twisted.internet.defer import inlineCallbacks, succeed

from ...base import TestBase
from ...mocking.endpoint import MockEndpointListener
from ...mocking.exit_socket import MockTunnelExitSocket
from ...mocking.ipv8 import MockIPv8
from ....messaging.anonymization.community import TunnelCommunity, TunnelSettings
from ....messaging.anonymization.endpoint import TunnelEndpoint
from ....messaging.anonymization.tunnel import CIRCUIT_STATE_EXTENDING, CIRCUIT_TYPE_IPV8, \
    PEER_FLAG_EXIT_ANY, PEER_FLAG_EXIT_IPV8
from ....messaging.interfaces.udp.endpoint import UDPEndpoint
from ....util import cast_to_bin


# Map of info_hash -> peer list
global_dht_services = defaultdict(list)


class MockDHTProvider(object):

    def __init__(self, peer):
        self.peer = peer
        # DHTDiscoveryCommunity functionality
        global_dht_services[peer.mid].append(peer)

    def peer_lookup(self, mid, cb):
        self.lookup(mid, cb)

    def lookup(self, info_hash, cb):
        if info_hash in global_dht_services:
            cb((info_hash, global_dht_services[info_hash]))

    def announce(self, info_hash, intro_point):
        global_dht_services[info_hash].append(intro_point)


class TestTunnelCommunity(TestBase):

    def setUp(self):
        super(TestTunnelCommunity, self).setUp()
        self.initialize(TunnelCommunity, 2)

        # An actual UDPEndpoint, if needed by the test (for catching exited data)
        self.public_endpoint = None

    def tearDown(self):
        # If an endpoint was used, close it
        if self.public_endpoint:
            self.public_endpoint.close()

        return super(TestTunnelCommunity, self).tearDown()

    def create_node(self):
        # Initialize a TunnelCommunity without circuits or exit node functionality
        settings = TunnelSettings()
        settings.min_circuits = 0
        settings.max_circuits = 0
        settings.remove_tunnel_delay = 0
        ipv8 = MockIPv8(u"curve25519", TunnelCommunity, settings=settings)
        # Then kill all automated circuit creation
        ipv8.overlay.cancel_all_pending_tasks()
        # Finally, use the proper exitnode and circuit settings for manual creation
        ipv8.overlay.settings.min_circuits = 1
        ipv8.overlay.settings.max_circuits = 1
        ipv8.overlay.dht_provider = MockDHTProvider(ipv8.overlay.my_peer)
        return ipv8

    def assert_no_more_tunnels(self):
        """
        Utility method to check whether there are no more tunnels left
        """
        for node in self.nodes:
            self.assertFalse(node.overlay.exit_sockets)
            self.assertFalse(node.overlay.relay_from_to)
            self.assertFalse(node.overlay.circuits)

    @inlineCallbacks
    def test_introduction_as_exit(self):
        """
        Check if introduction requests share the fact that nodes are exit nodes.
        """
        self.nodes[0].overlay.settings.peer_flags |= PEER_FLAG_EXIT_ANY

        yield self.introduce_nodes()

        self.assertIn(self.nodes[0].my_peer, self.nodes[1].overlay.get_candidates(PEER_FLAG_EXIT_ANY))
        self.assertNotIn(self.nodes[1].my_peer, self.nodes[0].overlay.get_candidates(PEER_FLAG_EXIT_ANY))

    @inlineCallbacks
    def test_introduction_as_exit_twoway(self):
        """
        Check if two nodes can have each other as exit nodes.
        """
        self.nodes[0].overlay.settings.peer_flags |= PEER_FLAG_EXIT_ANY
        self.nodes[1].overlay.settings.peer_flags |= PEER_FLAG_EXIT_ANY

        yield self.introduce_nodes()

        self.assertIn(self.nodes[0].my_peer, self.nodes[1].overlay.get_candidates(PEER_FLAG_EXIT_ANY))
        self.assertIn(self.nodes[1].my_peer, self.nodes[0].overlay.get_candidates(PEER_FLAG_EXIT_ANY))

    @inlineCallbacks
    def test_introduction_as_exit_noway(self):
        """
        Check if two nodes don't advertise themselves as exit node incorrectly.
        """
        yield self.introduce_nodes()

        self.assertEqual(len(self.nodes[0].overlay.get_candidates(PEER_FLAG_EXIT_ANY)), 0)
        self.assertEqual(len(self.nodes[1].overlay.get_candidates(PEER_FLAG_EXIT_ANY)), 0)

    @inlineCallbacks
    def test_create_circuit(self):
        """
        Check if 1 hop circuit creation works.
        """
        self.nodes[1].overlay.settings.peer_flags |= PEER_FLAG_EXIT_ANY
        yield self.introduce_nodes()

        # Let node 0 build tunnels of 1 hop (settings.min_circuits = settings.max_circuits = 1)
        # It should use node 1 for this
        self.nodes[0].overlay.build_tunnels(1)

        # Let the circuit creation commence
        yield self.deliver_messages()

        # Node 0 should now have all of its required 1 hop circuits (1.0/100%)
        self.assertEqual(self.nodes[0].overlay.tunnels_ready(1), 1.0)
        # Node 1 has an exit socket open
        self.assertEqual(len(self.nodes[1].overlay.exit_sockets), 1)

    @inlineCallbacks
    def test_create_circuit_no_exit(self):
        """
        Check if 1 hop circuit creation fails without exit nodes.
        """
        yield self.introduce_nodes()
        self.nodes[0].overlay.build_tunnels(1)

        # Attempt circuit creation
        yield self.deliver_messages()

        # Node 0 should now have no 1 hop circuits (0.0/0%)
        self.assertEqual(self.nodes[0].overlay.tunnels_ready(1), 0.0)
        # Node 1 should not have an exit socket open
        self.assertEqual(len(self.nodes[1].overlay.exit_sockets), 0)

    @inlineCallbacks
    def test_create_circuit_multiple_calls(self):
        """
        Check if circuit creation is aborted when it's already building the requested circuit.
        """
        self.nodes[1].overlay.settings.peer_flags |= PEER_FLAG_EXIT_ANY
        yield self.introduce_nodes()

        # Don't allow the exit node to answer, this keeps peer 0's circuit in EXTENDING state
        self.nodes[1].overlay.endpoint.close()
        self.nodes[0].overlay.build_tunnels(1)

        # Node 0 should have 1 circuit in the CIRCUIT_STATE_EXTENDING state
        self.assertEqual(len(self.nodes[0].overlay.find_circuits(state=CIRCUIT_STATE_EXTENDING)), 1)

        # Subsequent calls to build_circuits should not change this
        self.nodes[0].overlay.build_tunnels(1)
        self.assertEqual(len(self.nodes[0].overlay.find_circuits(state=CIRCUIT_STATE_EXTENDING)), 1)

    @inlineCallbacks
    def test_destroy_circuit_from_originator(self):
        """
        Check if a 2 hop circuit can be destroyed (by the exit node)
        """
        self.add_node_to_experiment(self.create_node())
        self.nodes[2].overlay.settings.peer_flags |= PEER_FLAG_EXIT_ANY
        yield self.introduce_nodes()
        self.nodes[0].overlay.build_tunnels(2)
        yield self.deliver_messages()

        # Destroy the circuit we just created using a destroy message
        yield self.nodes[0].overlay.remove_circuit(list(self.nodes[0].overlay.circuits.keys())[0], destroy=True)
        yield self.deliver_messages()

        self.assert_no_more_tunnels()

    @inlineCallbacks
    def test_destroy_circuit_from_exit(self):
        """
        Check if a 2 hop circuit can be destroyed (by the exit node)
        """
        self.add_node_to_experiment(self.create_node())
        self.nodes[2].overlay.settings.peer_flags |= PEER_FLAG_EXIT_ANY
        yield self.introduce_nodes()
        self.nodes[0].overlay.build_tunnels(2)
        yield self.deliver_messages()

        self.nodes[2].overlay.remove_exit_socket(list(self.nodes[2].overlay.exit_sockets.keys())[0], destroy=True)
        yield self.deliver_messages()

        self.assert_no_more_tunnels()

    @inlineCallbacks
    def test_destroy_circuit_from_relay(self):
        """
        Check if a 2 hop circuit can be destroyed (by the relay node)
        """
        self.add_node_to_experiment(self.create_node())
        self.nodes[2].overlay.settings.peer_flags |= PEER_FLAG_EXIT_ANY
        yield self.introduce_nodes()
        self.nodes[0].overlay.build_tunnels(2)
        yield self.deliver_messages()

        self.nodes[1].overlay.remove_relay(list(self.nodes[1].overlay.relay_from_to.keys())[0], destroy=True)
        yield self.deliver_messages()

        self.assert_no_more_tunnels()

    @inlineCallbacks
    def test_destroy_circuit_bad_id(self):
        """
        Check if the correct circuit gets destroyed.
        """
        self.nodes[1].overlay.settings.peer_flags |= PEER_FLAG_EXIT_ANY
        yield self.introduce_nodes()
        self.nodes[0].overlay.build_tunnels(1)
        yield self.deliver_messages()

        # Destroy a circuit which does not exist (circuit_id + 1)
        # This should not affect other circuits
        yield self.nodes[0].overlay.remove_circuit(list(self.nodes[0].overlay.circuits.keys())[0] + 1, destroy=True)
        yield self.deliver_messages()

        # Node 0 should still have all of its required 1 hop circuits (1.0/100%)
        self.assertEqual(self.nodes[0].overlay.tunnels_ready(1), 1.0)
        # Node 1 still has an exit socket open
        self.assertEqual(len(self.nodes[1].overlay.exit_sockets), 1)

    @inlineCallbacks
    def test_tunnel_data(self):
        """
        Check if data is correctly exited.
        """
        # Listen in on communication of the target
        self.public_endpoint = UDPEndpoint(8080)
        self.public_endpoint.open()
        ep_listener = MockEndpointListener(self.public_endpoint)

        # Build a tunnel
        self.nodes[1].overlay.settings.peer_flags |= PEER_FLAG_EXIT_ANY
        yield self.introduce_nodes()
        self.nodes[0].overlay.build_tunnels(1)
        yield self.deliver_messages()

        # Construct a data packet
        prefix = b'\x00' * 23
        data = prefix + cast_to_bin(''.join([chr(i) for i in range(256)]))

        self.public_endpoint.assert_open()

        # Tunnel the data to the endpoint
        circuit = list(self.nodes[0].overlay.circuits.values())[0]
        self.nodes[0].overlay.send_data([circuit.peer.address], circuit.circuit_id,
                                        ('localhost', self.public_endpoint.get_address()[1]), ('0.0.0.0', 0), data)
        # This is not test communication, but actual socket communication, we can't do a smart sleep
        yield self.sleep()

        self.assertEqual(len(ep_listener.received_packets), 1)
        self.assertEqual(ep_listener.received_packets[0][1], data)

    @inlineCallbacks
    def test_two_hop_circuit(self):
        """
        Check if a two hop circuit is correctly created.

        Note that we avoid exit nodes in the relay path, so we explicitly set relay nodes to not be exits.
        """
        self.add_node_to_experiment(self.create_node())

        # Build a tunnel
        self.nodes[1].overlay.settings.peer_flags |= PEER_FLAG_EXIT_ANY
        yield self.introduce_nodes()
        self.nodes[0].overlay.build_tunnels(2)
        yield self.deliver_messages()

        self.assertEqual(self.nodes[0].overlay.tunnels_ready(2), 1.0)

    @inlineCallbacks
    def test_three_hop_circuit(self):
        """
        Check if a three hop circuit is correctly created.

        Note that we avoid exit nodes in the relay path, so we explicitly set relay nodes to not be exits.
        """
        self.add_node_to_experiment(self.create_node())
        self.add_node_to_experiment(self.create_node())

        # Build a tunnel
        self.nodes[1].overlay.settings.peer_flags |= PEER_FLAG_EXIT_ANY
        yield self.introduce_nodes()
        self.nodes[0].overlay.build_tunnels(3)
        yield self.deliver_messages()

        self.assertEqual(self.nodes[0].overlay.tunnels_ready(3), 1.0)

    @inlineCallbacks
    def test_create_two_circuit(self):
        """
        Check if multiple 1 hop circuit creation works.
        """
        self.add_node_to_experiment(self.create_node())
        self.nodes[0].overlay.settings.min_circuits = 2
        self.nodes[0].overlay.settings.max_circuits = 2
        self.nodes[1].overlay.settings.peer_flags |= PEER_FLAG_EXIT_ANY
        self.nodes[2].overlay.settings.peer_flags |= PEER_FLAG_EXIT_ANY
        yield self.introduce_nodes()

        # Let node 0 build tunnels of 1 hop (settings.min_circuits = settings.max_circuits = 2)
        # It should use node 1 and 2 for this
        self.nodes[0].overlay.build_tunnels(1)

        # Let the circuit creation commence
        yield self.deliver_messages()

        # Node 0 should now have all of its required 1 hop circuits (1.0/100%)
        self.assertEqual(self.nodes[0].overlay.tunnels_ready(1), 1.0)
        self.assertEqual(len(self.nodes[0].overlay.circuits), 2)
        # Two exit sockets are open between node 1 and 2 (NOT evenly spread)
        self.assertEqual(len(self.nodes[1].overlay.exit_sockets) + len(self.nodes[2].overlay.exit_sockets), 2)

    @inlineCallbacks
    def test_reuse_partial_circuit(self):
        """
        Check if we can change the unverified hop of a circuit.
        """
        self.add_node_to_experiment(self.create_node())

        self.nodes[2].overlay.settings.peer_flags |= PEER_FLAG_EXIT_ANY
        self.nodes[2].overlay.should_join_circuit = lambda *args: succeed(False)
        yield self.introduce_nodes()
        self.nodes[0].overlay.build_tunnels(2)
        yield self.deliver_messages()

        # We wanted to create circuit 0 -> 1 -> 2, but node 2 is not responding
        circuit = list(self.nodes[0].overlay.circuits.values())[0]
        self.assertEqual([h.mid for h in circuit.hops], [self.nodes[1].overlay.my_peer.mid])
        self.assertEqual(circuit.unverified_hop.mid, self.nodes[2].overlay.my_peer.mid)

        # Let's add a new exit node, and retry to extend the circuit
        self.add_node_to_experiment(self.create_node())
        self.nodes[3].overlay.settings.peer_flags |= PEER_FLAG_EXIT_ANY
        yield self.introduce_nodes()
        # Let's pretend that node 1 selected node 3 as a possible node for circuit extension
        cache = self.nodes[1].overlay.request_cache.get(u"created", circuit.circuit_id)
        cache.candidates[self.nodes[3].overlay.my_peer.public_key.key_to_bin()] = self.nodes[3].overlay.my_peer

        # Retry to extend the circuit
        circuit.required_exit = None
        self.nodes[0].overlay.send_extend(circuit, [self.nodes[3].overlay.my_peer.public_key.key_to_bin()], 1)
        yield self.deliver_messages()

        # Circuit should now be 0 -> 1 -> 3
        self.assertEqual([h.mid for h in circuit.hops], [self.nodes[1].overlay.my_peer.mid,
                                                         self.nodes[3].overlay.my_peer.mid])
        self.assertEqual(circuit.unverified_hop, None)
        self.assertEqual(self.nodes[0].overlay.tunnels_ready(2), 1.0)

    @inlineCallbacks
    def test_reuse_partial_circuit_first_hop(self):
        """
        Check if we can change the first unverified hop of a circuit.
        """
        self.add_node_to_experiment(self.create_node())

        self.nodes[2].overlay.settings.peer_flags |= PEER_FLAG_EXIT_ANY
        self.nodes[1].overlay.should_join_circuit = lambda *args: succeed(False)
        yield self.introduce_nodes()
        self.nodes[0].overlay.build_tunnels(2)
        yield self.deliver_messages()

        # Adding the first hop fails, since hop 1 is not responding
        circuit = list(self.nodes[0].overlay.circuits.values())[0]
        self.assertEqual(circuit.hops, ())
        self.assertEqual(circuit.unverified_hop.mid, self.nodes[1].overlay.my_peer.mid)

        # Let's add a new node, and retry to extend the circuit
        self.add_node_to_experiment(self.create_node())
        yield self.introduce_nodes()

        # Retry to extend the circuit
        self.nodes[0].overlay.send_initial_create(circuit, [self.nodes[3].overlay.my_peer], 1)
        yield self.deliver_messages()

        # Circuit should now be 0 -> 2 -> 3
        self.assertEqual([h.mid for h in circuit.hops], [self.nodes[3].overlay.my_peer.mid,
                                                         self.nodes[2].overlay.my_peer.mid])
        self.assertEqual(circuit.unverified_hop, None)
        self.assertEqual(self.nodes[0].overlay.tunnels_ready(2), 1.0)

    @inlineCallbacks
    def test_tunnel_endpoint_anon(self):
        """
        Check if the tunnel endpoint is routing traffic correctly with anonymity enabled.
        """
        self.add_node_to_experiment(self.create_node())
        self.nodes[2].overlay.settings.peer_flags |= PEER_FLAG_EXIT_IPV8
        yield self.introduce_nodes()
        self.nodes[0].overlay.create_circuit(1, CIRCUIT_TYPE_IPV8)
        yield self.deliver_messages()

        exit_socket = list(self.nodes[2].overlay.exit_sockets.values())[0]
        self.nodes[2].overlay.exit_sockets[exit_socket.circuit_id] = MockTunnelExitSocket(exit_socket)

        sender = self.nodes[0].overlay
        send_data_org = sender.send_data

        def send_data(*args):
            sender.called = True
            send_data_org(*args)
        sender.called = False
        sender.send_data = send_data

        prefix = b'\x00\x01' + b'\x00' * 20
        self.nodes[0].overlay.endpoint = endpoint = TunnelEndpoint(self.nodes[0].overlay.endpoint)
        endpoint.set_tunnel_community(self.nodes[0].overlay)
        endpoint.set_anonymity(prefix, True)

        ep_listener = MockEndpointListener(self.nodes[1].endpoint)
        endpoint.send(self.nodes[1].overlay.my_estimated_wan, prefix + b'DATA')
        yield self.deliver_messages()
        self.assertEqual(len(ep_listener.received_packets), 1)
        self.assertEqual(ep_listener.received_packets[0][1], prefix + b'DATA')
        self.assertTrue(sender.called)

        # When a circuit closes, sending data should fail
        sender.called = False
        circuit = self.nodes[0].overlay.find_circuits(ctype=CIRCUIT_TYPE_IPV8)[0]
        self.nodes[0].overlay.remove_circuit(circuit.circuit_id)
        endpoint.send(self.nodes[1].overlay.my_estimated_wan, prefix + b'DATA')
        yield self.deliver_messages()
        self.assertEqual(len(ep_listener.received_packets), 1)
        self.assertFalse(sender.called)

    @inlineCallbacks
    def test_tunnel_endpoint_no_anon(self):
        """
        Check if the tunnel endpoint is routing traffic correctly with anonymity disabled.
        """

        prefix = b'\x00' * 22
        self.nodes[0].overlay.endpoint = endpoint = TunnelEndpoint(self.nodes[0].overlay.endpoint)
        endpoint.set_tunnel_community(self.nodes[0].overlay)
        endpoint.set_anonymity(prefix, False)

        ep_listener = MockEndpointListener(self.nodes[1].endpoint)
        endpoint.send(self.nodes[1].overlay.my_estimated_wan, prefix + b'DATA')
        yield self.deliver_messages()

        self.assertEqual(len(ep_listener.received_packets), 1)
        self.assertEqual(ep_listener.received_packets[0][1], prefix + b'DATA')



from __future__ import absolute_import

import time
import hashlib

from collections import defaultdict


class Value(object):
    """
    Class for storing DHT values.
    """

    def __init__(self, id_, data, max_age, version):
        self.id = id_
        self.data = data
        self.last_update = time.time()
        self.max_age = max_age
        self.version = version

    @property
    def age(self):
        return time.time() - self.last_update

    @property
    def expired(self):
        return self.age > self.max_age

    def __eq__(self, other):
        return self.id == other.id

    def __hash__(self):
        return 0


class Storage(object):
    """
    Class for storing key-value pairs in memory.
    """

    def __init__(self):
        self.items = defaultdict(list)

    def put(self, key, data, id_=None, max_age=86400, version=0):
        id_ = id_ or hashlib.sha1(data).digest()
        new_value = Value(id_, data, max_age, version)

        try:
            index = self.items[key].index(new_value)
            old_value = self.items[key][index]
            if new_value.version >= old_value.version:
                del self.items[key][index]
                self.items[key].insert(0, new_value)
                self.items[key].sort(key=lambda v: 1 if v.id == key else 0)
        except ValueError:
            self.items[key].insert(0, new_value)
            self.items[key].sort(key=lambda v: 1 if v.id == key else 0)

    def get(self, key, limit=None):
        return [value.data for value in self.items[key][:limit]] if key in self.items else []

    def items_older_than(self, min_age):
        items = []
        for key in self.items:
            items += [(key, value.data) for value in self.items[key] if value.age > min_age]
        return items

    def clean(self):
        for key in self.items:
            for index, value in reversed(list(enumerate(self.items[key]))):
                if value.expired:
                    del self.items[key][index]
                else:
                    break



from __future__ import absolute_import

import json
import os
from base64 import decodestring, encodestring
from binascii import unhexlify
from hashlib import sha1
from random import choice
from threading import Lock

from twisted.internet.defer import inlineCallbacks

from .bonehexact.algorithm import BonehExactAlgorithm
from .caches import (HashCache, PeerCache, PendingChallengeCache, ProvingAttestationCache,
                     ReceiveAttestationRequestCache, ReceiveAttestationVerifyCache)
from .database import AttestationsDB
from .payload import (AttestationChunkPayload, ChallengePayload, ChallengeResponsePayload, RequestAttestationPayload,
                      VerifyAttestationRequestPayload)
from .pengbaorange.algorithm import PengBaoRangeAlgorithm
from ..identity_formats import FORMATS
from ...community import Community
from ...lazy_community import lazy_wrapper
from ...messaging.payload_headers import BinMemberAuthenticationPayload, GlobalTimeDistributionPayload
from ...peer import Peer
from ...requestcache import RequestCache
from ...util import cast_to_bin, cast_to_chr, cast_to_unicode


ID_ALGORITHMS = {
    "bonehexact": BonehExactAlgorithm,
    "pengbaorange": PengBaoRangeAlgorithm
}

receive_block_lock = Lock()


def synchronized(f):
    """
    Due to database inconsistencies, we can't allow multiple threads to handle a received_half_block at the same time.
    """
    def wrapper(self, *args, **kwargs):
        with receive_block_lock:
            return f(self, *args, **kwargs)
    return wrapper


class AttestationCommunity(Community):
    """
    Community for sharing Attestations.

    Note that the logic for giving out Attestations is in the TrustChain.
    """
    master_peer = Peer(unhexlify("3081a7301006072a8648ce3d020106052b810400270381920004057a009787f66ea54d5082ea2f56a8"
                                 "42488e319c14c98967c39286433233f769a73e9c894149cf9053a9a0c2548f07171df9c46c3bdb106a"
                                 "fa9e9a8a06926e0ec35871c91f2ab1a20651d0a7b5fda209a3500a09b630a193b281a266230472ef0c"
                                 "c0622c793dc18eed6c57d7bcd1eeca33e2e38277ea99c28d4c62f850f81b5eb3eb19fcb601747bd87a"
                                 "a0b04e360ae9"))

    def __init__(self, *args, **kwargs):
        working_directory = kwargs.pop('working_directory', '')
        db_name = kwargs.pop('db_name', 'attestations')

        super(AttestationCommunity, self).__init__(*args, **kwargs)

        self.attestation_request_callback = lambda peer, attribute_name, metadata: None
        self.attestation_request_complete_callback = \
            lambda for_peer, attribute_name, attr_hash, id_format, from_peer=None: None
        self.verify_request_callback = lambda attribute_name, attr_hash: True

        # Map of attestation hash -> (PrivateKey, id_format)
        self.attestation_keys = {}
        self.database = AttestationsDB(working_directory, db_name)
        for attribute_hash, _, key, id_format in self.database.get_all():
            attribute_hash = attribute_hash if isinstance(attribute_hash, bytes) else str(attribute_hash)
            key = key if isinstance(key, bytes) else str(key)
            id_format = (id_format if isinstance(id_format, bytes) else str(id_format)).decode('utf-8')
            self.attestation_keys[attribute_hash] = (self.get_id_algorithm(id_format).load_secret_key(key), id_format)
        self.cached_attestation_blobs = {}

        self.request_cache = RequestCache()

        self.decode_map.update({
            chr(1): self.on_verify_attestation_request,
            chr(2): self.on_attestation_chunk,
            chr(3): self.on_challenge,
            chr(4): self.on_challenge_response,
            chr(5): self.on_request_attestation
        })

    def unload(self):
        self.request_cache.shutdown()

        super(AttestationCommunity, self).unload()
        # Close the database after we stop accepting requests.
        self.database.close()

    def get_id_algorithm(self, id_format):
        return ID_ALGORITHMS[FORMATS[id_format]["algorithm"]](id_format)

    def set_attestation_request_callback(self, f):
        """
        Set the callback to be called when someone requests an attestation from us.

        f should accept a (Peer, attribute name, metadata) and return a str()-able value.
        If it f returns None, no attestation is made.

        :param f: the callback function providing the value
        """
        self.attestation_request_callback = f

    def set_attestation_request_complete_callback(self, f):
        """
        f should accept a (Peer, attribute_name, hash, id_format, Peer=None), it is called when an Attestation
        has been made for another peer

        :param f: the function to call when an Attestation has been completed
        """
        self.attestation_request_complete_callback = f

    def set_verify_request_callback(self, f):
        """
        Set the callback to be called when someone wants to verify our attribute.

        f should accept a (Peer, attribute_name) and return a boolean value.
        If f return True, the attribute will be verified.

        :param f: the function to call when our attribute is requested for verification
        """
        self.verify_request_callback = f

    def request_attestation(self, peer, attribute_name, secret_key, metadata={}):
        """
        Request attestation of one of our attributes.

        :param peer: Peer of the Attestor
        :param attribute_name: the attribute we want attested
        :param secret_key: the secret key we use for this attribute
        """
        public_key = secret_key.public_key()
        id_format = metadata.pop("id_format", "id_metadata")
        self.request_cache.add(ReceiveAttestationRequestCache(self, peer.mid, secret_key, attribute_name, id_format))

        meta_dict = {
            "attribute": attribute_name,
            "public_key": cast_to_chr(encodestring(public_key.serialize())),
            "id_format": id_format
        }
        meta_dict.update(metadata)
        metadata = cast_to_bin(json.dumps({cast_to_unicode(k): cast_to_unicode(v) for k, v in meta_dict.items()}))

        global_time = self.claim_global_time()
        auth = BinMemberAuthenticationPayload(self.my_peer.public_key.key_to_bin()).to_pack_list()
        payload = RequestAttestationPayload(metadata).to_pack_list()
        dist = GlobalTimeDistributionPayload(global_time).to_pack_list()

        packet = self._ez_pack(self._prefix, 5, [auth, dist, payload])
        self.endpoint.send(peer.address, packet)

    @inlineCallbacks
    @lazy_wrapper(GlobalTimeDistributionPayload, RequestAttestationPayload)
    def on_request_attestation(self, peer, dist, payload):
        """
        Someone wants us to attest their attribute.
        """
        metadata = json.loads(payload.metadata)
        attribute = metadata.pop('attribute')
        pubkey_b64 = cast_to_bin(metadata.pop('public_key'))
        id_format = metadata.pop('id_format')
        id_algorithm = self.get_id_algorithm(id_format)

        value = yield self.attestation_request_callback(peer, attribute, metadata)
        if value is None:
            return

        PK = id_algorithm.load_public_key(decodestring(pubkey_b64))
        attestation_blob = id_algorithm.attest(PK, value)
        attestation = id_algorithm.get_attestation_class().unserialize(attestation_blob, id_format)

        self.attestation_request_complete_callback(peer, attribute, attestation.get_hash(), id_format)

        self.send_attestation(peer.address, attestation_blob)

    def on_attestation_complete(self, unserialized, secret_key, peer, name, attestation_hash, id_format):
        """
        We got an Attestation delivered to us.
        """
        self.attestation_keys[cast_to_bin(attestation_hash)] = (secret_key, id_format)
        self.database.insert_attestation(unserialized, attestation_hash, secret_key, id_format)
        self.attestation_request_complete_callback(self.my_peer, name, attestation_hash, id_format, peer)

    def verify_attestation_values(self, socket_address, attestation_hash, values, callback, id_format):
        """
        Ask the peer behind a socket address to deliver the Attestation with a certain hash.

        :param socket_address: the socket address to send to
        :param attestation_hash: the hash of the Attestation to request
        :param values: the values for which we want to measure certainty
        :param callback: the callback to call with the map of (hash, {value: certainty})
        :param id_format: the identity format specifier
        """
        algorithm = self.get_id_algorithm(id_format)

        def on_complete(attestation_hash, relativity_map):
            callback(attestation_hash, [algorithm.certainty(value, relativity_map) for value in values])
        self.request_cache.add(ProvingAttestationCache(self, attestation_hash, id_format, on_complete=on_complete))
        self.create_verify_attestation_request(socket_address, attestation_hash, id_format)

    def create_verify_attestation_request(self, socket_address, attestation_hash, id_format):
        """
        Ask the peer behind a socket address to deliver the Attestation with a certain hash.

        :param socket_address: the socket address to send to
        :param attestation_hash: the hash of the Attestation to request
        :param id_format: the identity format specifier
        """
        self.request_cache.add(ReceiveAttestationVerifyCache(self, attestation_hash, id_format))

        global_time = self.claim_global_time()
        auth = BinMemberAuthenticationPayload(self.my_peer.public_key.key_to_bin()).to_pack_list()
        payload = VerifyAttestationRequestPayload(attestation_hash).to_pack_list()
        dist = GlobalTimeDistributionPayload(global_time).to_pack_list()

        packet = self._ez_pack(self._prefix, 1, [auth, dist, payload])
        self.endpoint.send(socket_address, packet)

    @inlineCallbacks
    @lazy_wrapper(GlobalTimeDistributionPayload, VerifyAttestationRequestPayload)
    def on_verify_attestation_request(self, peer, dist, payload):
        """
        We received a request to verify one of our attestations. Send the requested attestation back.
        """
        attestation_blob, = self.database.get_attestation_by_hash(payload.hash)
        if not attestation_blob:
            return

        value = yield self.verify_request_callback(peer, payload.hash)
        if not value:
            return

        SK, id_format = self.attestation_keys[payload.hash]
        attestation_cls = self.get_id_algorithm(id_format).get_attestation_class()
        private_attestation = attestation_cls.unserialize_private(SK, attestation_blob, id_format)
        public_attestation_blob = private_attestation.serialize()
        self.cached_attestation_blobs[payload.hash] = private_attestation

        self.send_attestation(peer.address, public_attestation_blob)

    def send_attestation(self, socket_address, blob):
        # If we want to serve this request send the attestation in chunks of 800 bytes
        sequence_number = 0
        for i in range(0, len(blob), 800):
            blob_chunk = blob[i:i + 800]
            self.logger.debug("Sending attestation chunk %d to %s", sequence_number, str(socket_address))
            global_time = self.claim_global_time()
            auth = BinMemberAuthenticationPayload(self.my_peer.public_key.key_to_bin()).to_pack_list()
            payload = AttestationChunkPayload(sha1(blob).digest(), sequence_number, blob_chunk).to_pack_list()
            dist = GlobalTimeDistributionPayload(global_time).to_pack_list()
            packet = self._ez_pack(self._prefix, 2, [auth, dist, payload])
            self.endpoint.send(socket_address, packet)

            sequence_number += 1

    @lazy_wrapper(GlobalTimeDistributionPayload, AttestationChunkPayload)
    def on_attestation_chunk(self, peer, dist, payload):
        """
        We received a chunk of an Attestation.
        """
        hash_id = HashCache.id_from_hash(u"receive-verify-attestation", payload.hash)
        peer_id = PeerCache.id_from_address(u"receive-request-attestation", peer.mid)
        if self.request_cache.has(*hash_id):
            cache = self.request_cache.get(*hash_id)
            cache.attestation_map |= {(payload.sequence_number, payload.data), }

            serialized = b""
            for (_, chunk) in sorted(cache.attestation_map, key=lambda item: item[0]):
                serialized += chunk

            attestation_class = self.get_id_algorithm(cache.id_format).get_attestation_class()
            if sha1(serialized).digest() == payload.hash:
                unserialized = attestation_class.unserialize(serialized, cache.id_format)
                self.request_cache.pop(*hash_id)
                self.on_received_attestation(peer, unserialized, payload.hash)

            self.logger.debug("Received attestation chunk %d for proving by %s", payload.sequence_number, str(peer))
        elif self.request_cache.has(*peer_id):
            cache = self.request_cache.get(*peer_id)
            cache.attestation_map |= {(payload.sequence_number, payload.data), }

            serialized = b""
            for (_, chunk) in sorted(cache.attestation_map, key=lambda item: item[0]):
                serialized += chunk

            attestation_class = self.get_id_algorithm(cache.id_format).get_attestation_class()
            if sha1(serialized).digest() == payload.hash:
                unserialized = attestation_class.unserialize_private(cache.key, serialized, cache.id_format)
                cache = self.request_cache.pop(*peer_id)
                self.on_attestation_complete(unserialized, cache.key, peer, cache.name, unserialized.get_hash(),
                                             cache.id_format)

            self.logger.debug("Received attestation chunk %d for my attribute %s", payload.sequence_number, cache.name)
        else:
            self.logger.warning("Received Attestation chunk which we did not request!")

    def on_received_attestation(self, peer, attestation, attestation_hash):
        """
        Callback for when we got the entire attestation from a peer.

        :param peer: the Peer we got this attestation from
        :param attestation: the Attestation object we can check
        """
        algorithm = self.get_id_algorithm(attestation.id_format)

        relativity_map = algorithm.create_certainty_aggregate(attestation)
        hashed_challenges = []
        cache = self.request_cache.get(*HashCache.id_from_hash(u"proving-attestation", attestation_hash))
        cache.public_key = attestation.PK
        challenges = algorithm.create_challenges(attestation.PK, attestation)
        for challenge in challenges:
            challenge_hash = sha1(challenge).digest()
            hashed_challenges.append(challenge_hash)
        cache.relativity_map = relativity_map
        cache.hashed_challenges = hashed_challenges
        cache.challenges = challenges
        self.logger.debug("Sending %d challenges to %s", len(challenges), str(peer))
        remaining = 10
        for challenge in challenges:
            if remaining == 0:
                break
            elif self.request_cache.has(*PendingChallengeCache.id_from_hash(u"proving-hash", sha1(challenge).digest())):
                continue
            remaining -= 1
            self.request_cache.add(PendingChallengeCache(self, sha1(challenge).digest(), cache, cache.id_format))

            global_time = self.claim_global_time()
            auth = BinMemberAuthenticationPayload(self.my_peer.public_key.key_to_bin()).to_pack_list()
            payload = ChallengePayload(attestation_hash, challenge).to_pack_list()
            dist = GlobalTimeDistributionPayload(global_time).to_pack_list()

            packet = self._ez_pack(self._prefix, 3, [auth, dist, payload])
            self.endpoint.send(peer.address, packet)

    @lazy_wrapper(GlobalTimeDistributionPayload, ChallengePayload)
    def on_challenge(self, peer, dist, payload):
        """
        We received a challenge for an Attestation.
        """
        SK, id_format = self.attestation_keys[payload.attestation_hash]
        challenge_hash = sha1(payload.challenge).digest()
        algorithm = self.get_id_algorithm(id_format)
        attestation = self.cached_attestation_blobs[payload.attestation_hash]

        global_time = self.claim_global_time()
        auth = BinMemberAuthenticationPayload(self.my_peer.public_key.key_to_bin()).to_pack_list()
        payload = ChallengeResponsePayload(challenge_hash,
                                           algorithm.create_challenge_response(SK, attestation, payload.challenge)
                                           ).to_pack_list()
        dist = GlobalTimeDistributionPayload(global_time).to_pack_list()

        packet = self._ez_pack(self._prefix, 4, [auth, dist, payload])
        self.endpoint.send(peer.address, packet)

    @synchronized
    @lazy_wrapper(GlobalTimeDistributionPayload, ChallengeResponsePayload)
    def on_challenge_response(self, peer, dist, payload):
        """
        We received a response to our challenge
        """
        cache = self.request_cache.get(*HashCache.id_from_hash(u"proving-hash", payload.challenge_hash))
        if cache:
            self.request_cache.pop(*HashCache.id_from_hash(u"proving-hash", payload.challenge_hash))
            proving_cache = cache.proving_cache
            pcache_prefix, pcache_id = HashCache.id_from_hash(u"proving-attestation", proving_cache.hash)
            challenge = None
            if payload.challenge_hash in proving_cache.hashed_challenges:
                proving_cache.hashed_challenges.remove(payload.challenge_hash)
                for challenge in proving_cache.challenges[:]:
                    if sha1(challenge).digest() == payload.challenge_hash:
                        proving_cache.challenges.remove(challenge)
                        break
            algorithm = self.get_id_algorithm(proving_cache.id_format)
            if cache.honesty_check < 0:
                algorithm.process_challenge_response(proving_cache.relativity_map, challenge, payload.response)
            elif not algorithm.process_honesty_challenge(cache.honesty_check, payload.response):
                self.logger.error("%s tried to cheat in the ZKP!", peer.address[0])
                # Liar, Completed
                if self.request_cache.has(pcache_prefix, pcache_id):
                    self.request_cache.pop(pcache_prefix, pcache_id)
                proving_cache.attestation_callbacks(proving_cache.hash, algorithm.create_certainty_aggregate(None))
            if len(proving_cache.hashed_challenges) == 0:
                self.logger.info("Completed attestation verification")
                # Completed
                if self.request_cache.has(pcache_prefix, pcache_id):
                    self.request_cache.pop(pcache_prefix, pcache_id)
                proving_cache.attestation_callbacks(proving_cache.hash, proving_cache.relativity_map)
            else:
                # Send another proving hash
                honesty_check = algorithm.honesty_check and (ord(os.urandom(1)[0:1]) < 38)
                honesty_check_byte = choice([0, 1, 2]) if honesty_check else -1
                challenge = None
                if honesty_check:
                    while not challenge or self.request_cache.has(*HashCache.id_from_hash(u"proving-hash",
                                                                                          sha1(challenge).digest())):
                        challenge = algorithm.create_honesty_challenge(proving_cache.public_key, honesty_check_byte)
                if (not honesty_check) or (challenge and self.request_cache.has(*HashCache.id_from_hash(u"proving-hash",
                                                                                sha1(challenge).digest()))):
                    honesty_check_byte = -1
                    challenge = None
                    for c in proving_cache.challenges:
                        if not self.request_cache.has(*HashCache.id_from_hash(u"proving-hash", sha1(c).digest())):
                            challenge = c
                            break
                    if not challenge:
                        self.logger.debug("No more bitpairs to challenge!")
                        return
                self.logger.debug("Sending challenge: %d (%d)", honesty_check_byte,
                                  len(proving_cache.hashed_challenges))
                self.request_cache.add(PendingChallengeCache(self, sha1(challenge).digest(), proving_cache,
                                                             cache.id_format, honesty_check_byte))

                global_time = self.claim_global_time()
                auth = BinMemberAuthenticationPayload(self.my_peer.public_key.key_to_bin()).to_pack_list()
                payload = ChallengePayload(proving_cache.hash, challenge).to_pack_list()
                dist = GlobalTimeDistributionPayload(global_time).to_pack_list()

                packet = self._ez_pack(self._prefix, 3, [auth, dist, payload])
                self.endpoint.send(peer.address, packet)



from __future__ import absolute_import
from __future__ import print_function

from collections import deque, namedtuple
from os import makedirs, path, rename
from random import randint
from shutil import rmtree
from subprocess import call
from sys import exit as _exit
from time import sleep, time

from six.moves import xrange
from twisted.internet import reactor
from twisted.internet.defer import Deferred, DeferredList

# Check if we are running from the root directory
# If not, modify our path so that we can import IPv8
try:
    import ipv8
    del ipv8
except ImportError:
    import sys
    sys.path.append(path.abspath(path.join(path.dirname(__file__), "..")))

from ipv8_service import IPv8
from ipv8.configuration import get_default_configuration
from ipv8.keyvault.crypto import default_eccrypto
from ipv8.overlay import Overlay
from ipv8.peer import Peer


test_results = {}
TestResult = namedtuple('TestResult', ['bytes_received', 'bytes_sent', 'packets_received', 'packets_sent'])
TestResultPair = namedtuple('TestResultPair', ['initiator', 'counterparty'])


class LoadOverlay(Overlay):

    def __init__(self, master_peer, my_peer, endpoint, network, window=1, packet_size=1000, packets=10000):
        """
        Create the overlay to pump through packets.

        :param window: the window size for packets (1 is synchronous)
        :param packet_size: the size of each packet
        :param packets: the number of packets to send in the experiment
        """
        super(LoadOverlay, self).__init__(master_peer, my_peer, endpoint, network)

        init_time = time()
        self.bytes_received = deque([(init_time, 0)])
        self.bytes_sent = deque([(init_time, 0)])
        self.packets_received = deque([(init_time, 0)])
        self.packets_sent = deque([(init_time, 0)])

        self.window = window
        self.packets = packets
        self.packet_content = "".join(chr(randint(0,255)) for _ in range(packet_size))
        self.packet_size = packet_size

        self.done = Deferred()

    def send(self, address):
        self.endpoint.send(address, self.packet_content)
        self.packets_sent.append((time(), self.packets_sent[-1][1] + 1))
        self.bytes_sent.append((time(), self.bytes_sent[-1][1] + self.packet_size))

    def on_packet(self, packet):
        source_address, data = packet
        self.packets_received.append((time(), self.packets_received[-1][1] + 1))
        self.bytes_received.append((time(), self.bytes_received[-1][1] + len(data)))
        self.window += 1

        if len(self.packets_received) < self.packets:
            while self.window > 0 and len(self.packets_sent) < self.packets:
                self.send(source_address)
                self.window -= 1
        elif not self.done.called:
            while self.window > 0 and len(self.packets_sent) < self.packets:
                self.send(source_address)
                self.window -= 1
            self.done.callback(self)

    def walk_to(self, address):
        pass

    def get_new_introduction(self, from_peer=None):
        return None

    def get_peers(self):
        return []


def setup_test(window=1, packet_size=1000, packets=10000):
    """
    Create two nodes who will be sending packets to each other.

    :param window: the window size for packets (1 is synchronous)
    :param packet_size: the size of each packet
    :param packets: the number of packets to send in the experiment
    :return: the deferred that fires once the experiment is complete
    """
    configuration = get_default_configuration()
    configuration['overlays'] = []
    configuration['keys'] = []

    master_peer = Peer(default_eccrypto.generate_key(u"low"))

    peer_ipv8 = IPv8(configuration)
    peer_ipv8.keys = {'my_peer': Peer(default_eccrypto.generate_key(u"low"))}
    peer_ipv8.overlays = [LoadOverlay(master_peer, peer_ipv8.keys['my_peer'], peer_ipv8.endpoint, peer_ipv8.network,
                                      window, packet_size, packets)]

    counterparty_ipv8 = IPv8(configuration)
    counterparty_ipv8.keys = {'my_peer': Peer(default_eccrypto.generate_key(u"low"))}
    counterparty_ipv8.overlays = [LoadOverlay(master_peer, counterparty_ipv8.keys['my_peer'],
                                              counterparty_ipv8.endpoint, counterparty_ipv8.network, 0,
                                              packet_size, packets)]

    peer_ipv8.overlays[0].send(counterparty_ipv8.endpoint.get_address())

    return DeferredList([peer_ipv8.overlays[0].done, counterparty_ipv8.overlays[0].done])


def on_synchronous_results(results):
    """
    We got the results of the first, synchronous, experiment.
    Start the next experiment.

    :param results: the two overlays
    """
    global test_results
    (_, overlay1), (__, overlay2) = results
    test_results["synchronous"] = TestResultPair(
        TestResult(overlay1.bytes_received, overlay1.bytes_sent, overlay1.packets_received, overlay1.packets_sent),
        TestResult(overlay2.bytes_received, overlay2.bytes_sent, overlay2.packets_received, overlay2.packets_sent)
    )
    test_asynchronous()


def on_asynchronous_results(results):
    """
    We got the results of the second, asynchronous, experiment.
    This is the last experiment, stop the reactor.

    :param results: the two overlays
    """
    global test_results
    (_, overlay1), (__, overlay2) = results
    test_results["asynchronous"] = TestResultPair(
        TestResult(overlay1.bytes_received, overlay1.bytes_sent, overlay1.packets_received, overlay1.packets_sent),
        TestResult(overlay2.bytes_received, overlay2.bytes_sent, overlay2.packets_received, overlay2.packets_sent)
    )
    reactor.callFromThread(reactor.stop)


def test_synchronous():
    """
    Start a synchronous test: 1kb packets, 100000 packets total.
    """
    setup_test(1, 1000, 20000).addCallback(on_synchronous_results)


def test_asynchronous():
    """
    Start an asynchronous test: 1kb packets, 100000 packets total.
    """
    setup_test(20, 1000, 20000).addCallback(on_asynchronous_results)


def run_tests():
    """
    Run the synchronous and asynchronous tests (one triggers the other).
    """
    test_synchronous()


# Start the tests and wait for them to finish.
reactor.callWhenRunning(run_tests)
reactor.run()

# Gather all the results.
prefix = "./endpoint_stresstest_results/"
old_prefix = None
try:
    makedirs(prefix)
except:
    old_prefix = path.abspath(prefix) + '_old'
    rmtree(old_prefix, ignore_errors=True)
    rename(path.abspath(prefix), old_prefix)
    makedirs(prefix)
for experiment, pair in test_results.items():
    for peer in ['initiator', 'counterparty']:
        for metric in ['bytes_received', 'bytes_sent', 'packets_received', 'packets_sent']:
            filename = prefix + "%s_%s_%s.csv" % (experiment, metric, peer)
            with open(filename, 'w') as f:
                f.write("time,count\n")
                value = getattr(getattr(pair, peer), metric)
                for time, count in value:
                    f.write("%f,%d\n" % (time, count))

# Allow the operating system to properly close the files and use R to plot.
sleep(2.0)
rcmd = "Rscript --vanilla %s \"%s\"" % (path.join(path.dirname(__file__), "endpoint_stresstest_plot.R"),
                                        path.abspath(prefix) + path.sep)
if old_prefix:
    rcmd += " \"%s\"" % (path.abspath(old_prefix) + path.sep)
retcode = call(rcmd, shell=True)

if retcode > 0:
    print("Error! Stresstest found significant slowdown:")
    binretcode = bin(retcode)[2:]
    errmap = {
        0: 'data_rcv_initiator',
        1: 'data_rcv_counterparty',
        2: 'data_snd_initiator',
        3: 'data_snd_counterparty',
        4: 'data_arcv_initiator',
        5: 'data_arcv_counterparty',
        6: 'data_asnd_initiator',
        7: 'data_asnd_counterparty'
    }
    for i in xrange(len(binretcode)):
        if binretcode[i] == '1':
            print("Significant slowdown in %s" % errmap[i])
else:
    print("Stresstest found no significant slowdown")

# This should be the last block of code in the file
if retcode > 0:
    _exit(1)
else:
    _exit(0)



"""
Implementation of the Boneh 2-DNF scheme ("Evaluating 2-DNF Formulas on Ciphertexts" by Boneh et al.).
"""
from __future__ import absolute_import
from __future__ import division

from random import randint

from .cryptography_wrapper import generate_safe_prime, is_prime
from .structs import BonehPrivateKey, BonehPublicKey
from .value import FP2Value
from .ec import weilpairing


def generate_prime(n):
    """
    Generate p = l * n - 1 such that:
     - p is "prime"
     - p mod 3 = 2
     - l is an as small as possible positive integer
    """
    p = 1
    l = 0
    while (p % 3) != 2 or not is_prime(p):
        l += 1
        p = l * n - 1
    return p


def bilinear_group(n, p, g1x, g1y, g2x, g2y):
    """
    Generate a bilinear group for two generators.
    """
    try:
        wp = weilpairing(p,
                         n,
                         (FP2Value(p, g1x), FP2Value(p, g1y)),
                         (FP2Value(p, b=g2x), FP2Value(p, g2y)),
                         (FP2Value(p), FP2Value(p)))
        return wp
    except:
        return FP2Value(p)


def get_random_exponentiation(p, n):
    """
    Create a random exponentiation of p in message space n.
    """
    r = randint(4, n - 1)
    test = p.intpow(r)
    while test == FP2Value(p.mod, 1):
        test = p.intpow(randint(4, n - 1))
    return test


def get_random_base(n):
    """
    Create a generator for the EC.
    """
    x = randint(2, n - 1)
    y = randint(2, n - 1)
    return x, y


def is_good_wp(n, wp):
    """
    A good pairing is not 0 and has order n.
    """
    is_one = wp == FP2Value(wp.mod, 1)
    is_zero = wp == FP2Value(wp.mod)
    good_order = wp.intpow(n + 1) == wp
    return good_order and not is_zero and not is_one


def get_good_wp(n, p=None):
    """
    Instead of inspecting torsion points and checking for co-primality:
    just brute force generate pairings until we a get a good one.
    :return: modulus, weilparing
    """
    wp = None
    if not p:
        p = generate_prime(n)
    while (wp is None) or not is_good_wp(n, wp):
        g1x, g1y = get_random_base(n)
        wp = bilinear_group(n, p, g1x, g1y, g1x, g1y)
        if not is_good_wp(n, wp):
            wp = wp.intpow((p + 1) // n)
    return p, wp


def generate_primes(key_size=128):
    """
    Generate some primes. Key size in bits.
    """
    if key_size >= 512:
        from cryptography.hazmat.backends import default_backend
        from cryptography.hazmat.primitives.asymmetric import rsa
        private_key = rsa.generate_private_key(public_exponent=65537, key_size=key_size, backend=default_backend())
        private_numbers = private_key.private_numbers()
        p, q = private_numbers.p, private_numbers.q
    else:
        p, q = generate_safe_prime(key_size), generate_safe_prime(key_size)
    return min(p, q), max(p, q)


def generate_keypair(key_size=32):
    """
    Generate a keypair for a certain prime bit space.
    """
    t1, t2 = generate_primes(key_size)
    n = t1 * t2
    p, g = get_good_wp(n)
    u = None
    while not u or (u.intpow(t2) == FP2Value(p, 1)):
        _, u = get_good_wp(n, p)
    h = u.intpow(t2)
    return BonehPublicKey(p, g, h), BonehPrivateKey(p, g, h, t1 * t2, t1)


def encode(pubkey, m):
    """
    Encode a message m given a public key.
    """
    return pubkey.g.intpow(m) * get_random_exponentiation(pubkey.h, pubkey.p)


def decode(privkey, msgspace, c):
    """
    Decode a ciphertext c given a private key and the possible source messages.
    """
    d = c.intpow(privkey.t1)
    t = privkey.g.intpow(privkey.t1)
    for m in msgspace:
        if d == t.intpow(m):
            return m
    return None



from __future__ import absolute_import

import errno
from collections import deque
import socket
import sys

from six.moves import xrange
from twisted.internet import protocol, reactor, error
from twisted.internet.error import MessageLengthError

from ..endpoint import Endpoint, EndpointClosedException

UDP_MAX_SIZE = 2 ** 16 - 60


class UDPEndpoint(Endpoint, protocol.DatagramProtocol):

    def __init__(self, port, ip="0.0.0.0"):
        Endpoint.__init__(self)
        self._port = port
        self._ip = ip
        self._running = False
        self._listening_port = False
        # If the outbound network buffer is blocked, buffer up to 100 packets
        # Pop from the left and append to the right side of the double-ended queue
        self._delayed_packets = deque(maxlen=min(100, max(0, sys.getrecursionlimit() - 2)))

        self.bytes_up = 0
        self.bytes_down = 0

    def datagramReceived(self, datagram, addr):
        self.bytes_down += len(datagram)
        self.notify_listeners((addr, datagram))

    def send(self, socket_address, packet):
        """
        Send a packet to a given address.
        :param socket_address: Tuple of (IP, port) which indicates the destination of the packet.
        :param packet: The packet to send.
        """
        self.assert_open()
        try:
            self.transport.write(packet, socket_address)
            self.bytes_up += len(packet)
            # If the write succeeded, try sending one of our previously blocked packets
            try:
                self.send(*self._delayed_packets.popleft())
            except IndexError:
                # This can happen if the queue is empty, in which case we don't care
                pass
        except socket.error as exc:
            # Not all OSes have WSAEWOULDBLOCK: Windows may have a blocked output buffer
            errnum = exc[0] if hasattr(exc, "__getitem__") else exc.errno
            if errnum == getattr(errno, 'WSAEWOULDBLOCK', 10035):
                self._logger.info("Rescheduling packet (due to blocked socket) outbound to %s", str(socket_address))
                self._delayed_packets.append((socket_address, packet))
            else:
                self._logger.warning("Dropping packet due to socket error: %s", exc)
        except MessageLengthError:
            self._logger.error("Sending a packet that is too big (length: %d)", len(packet))

    def open(self):
        for _ in xrange(10000):
            try:
                self._listening_port = reactor.listenUDP(self._port, self, self._ip, UDP_MAX_SIZE)
                self._logger.debug("Listening at %d", self._port)
                break
            except error.CannotListenError:
                self._logger.debug("Listening failed at %d", self._port)
                self._port += 1
                continue
        self._running = True
        return True

    def assert_open(self):
        if not self._running:
            raise EndpointClosedException(self)

    def close(self):
        self._running = False
        return self._listening_port.stopListening()

    def get_address(self):
        """
        Get the address for this Endpoint.
        """
        self.assert_open()
        return (self._listening_port.getHost().host,
                self._listening_port.getHost().port)

    def is_open(self):
        """
        Check if the underlying socket is open.
        """
        return self._listening_port and self._running



from __future__ import absolute_import

import inspect

from six import string_types
from six.moves import xrange

from .payload import Payload


class VariablePayload(Payload):
    """
    A Payload instance which mimics a struct. Useful for when you want a less verbose way to specify Payloads.

    This class requires you to only specify your format in:

        - <MyPayload>.format_list : a list of Serializer format specifications
        - <MyPayload>.names : the field names to use for the given formats

    For instance:

        class MyPayload(VariablePayload):
            format_list = ['?']
            names = ["is_this_a_boolean"]

    If you require field-specific pack/unpack operations you can specify them using the `fix_pack_*` and
    `fix_unpack_*` methods.
    Custom packing and unpacking rules can be useful for compression methods like socket.inet_aton, which you only
    want to apply when actually sending over the wire.
    """

    names = []

    def __init__(self, *args, **kwargs):
        """
        Instantiate this VariablePayload class.

        :param args: the anonymous list of arguments, an index-based mapping to self.names and self.format_list
        :param kwargs: the named arguments, mapping to self.names and self.format_list (in no particular order)
        :raises KeyError: if the given arguments do not match the class specification
        """
        index = 0
        fwd_args = {}
        # If our super class is an old-style Payload function, forward the required arguments.
        if not issubclass(super(VariablePayload, self).__class__, VariablePayload) and \
                inspect.ismethod(super(VariablePayload, self).__init__):
            super_argspec = inspect.getargspec(super(VariablePayload, self).__init__).args[1:]
            for arg in super_argspec:
                if arg in kwargs:
                    fwd_args[arg] = kwargs.pop(arg)
                else:
                    fwd_args[arg] = args[index]
                index += 1
            super(VariablePayload, self).__init__(**fwd_args)
        Payload.__init__(self)
        # Try to fill the required format specification.
        for _ in range(len(self.format_list) - index):
            # Run out all anonymous arguments, then start popping from the keyword arguments.
            # This will raise a KeyError if we provide more arguments than we can handle.
            value = args[index] if index < len(args) else kwargs.pop(self.names[index])
            setattr(self, self.names[index], value)
            index += 1
        if index != len(self.format_list):
            raise KeyError("%s missing %d arguments!" % (self.__class__.__name__, len(args) - index))
        # Try to fill in the optional format specification.
        for _ in self.optional_format_list:
            # If we run out of anonymous and named arguments, we stop.
            if index == len(args) and not kwargs:
                break
            value = args[index] if index < len(args) else kwargs.pop(self.names[index])
            setattr(self, self.names[index], value)
            index += 1
        if kwargs:
            raise KeyError("%s has leftover keyword arguments: %s!" % (self.__class__.__name__, str(kwargs)))

    @classmethod
    def from_unpack_list(cls, *args):
        """
        Given a list of raw arguments, initialize a new cls instance.

        If this class has special rules for certain fields, apply them.
        """
        unpack_args = list(args)
        for i in xrange(len(args)):
            custom_rule = "fix_unpack_" + cls.names[i]
            if hasattr(cls, custom_rule):
                unpack_args[i] = getattr(cls, custom_rule)(args[i])
        return cls(*unpack_args)

    @staticmethod
    def _to_packlist_fmt(fmt):
        return fmt if isinstance(fmt, string_types) else 'payload'

    def _fix_pack(self, name):
        """
        Check if there are custom rules for sending this field.
        """
        raw_value = getattr(self, name)
        custom_rule = "fix_pack_" + name
        if hasattr(self, custom_rule):
            return getattr(self, custom_rule)(raw_value)
        return raw_value

    def to_pack_list(self):
        """
        Convert the VariablePayload to a Serializable pack list.
        This method will automatically pull from the available format names and set instance fields.

        :return: the pack list
        """
        out = []
        index = 0
        for _ in range(len(self.format_list)):
            out.append((self._to_packlist_fmt(self.format_list[index]), self._fix_pack(self.names[index])))
            index += 1
        while index < len(self.names) and hasattr(self, self.names[index]):
            out.append((self._to_packlist_fmt(self.optional_format_list[index - len(self.format_list)]),
                        self._fix_pack(self.names[index])))
        index += 1
        return out



from __future__ import absolute_import

from twisted.trial import unittest

from ...keyvault.crypto import ECCrypto
from ...keyvault.keys import Key, PrivateKey, PublicKey
from ...keyvault.private.m2crypto import M2CryptoPK, M2CryptoSK
from ...keyvault.private.libnaclkey import LibNaCLPK, LibNaCLSK


class TestECCrypto(unittest.TestCase):

    m2crypto_key = ECCrypto().generate_key(u"very-low")
    libnacl_key = ECCrypto().generate_key(u"curve25519")

    def setUp(self):
        super(TestECCrypto, self).setUp()
        self.ecc = ECCrypto()

    def test_available(self):
        """
        Check if the required curves are available.
        """
        available = self.ecc.security_levels

        self.assertIn(u"very-low", available)
        self.assertIn(u"low", available)
        self.assertIn(u"medium", available)
        self.assertIn(u"high", available)
        self.assertIn(u"curve25519", available)

    def test_generate_m2crypto(self):
        """
        Check if M2Crypto backend keys can be generated correctly.
        """
        self.assertIsInstance(TestECCrypto.m2crypto_key, Key)
        self.assertIsInstance(TestECCrypto.m2crypto_key, PrivateKey)
        self.assertIsInstance(TestECCrypto.m2crypto_key, PublicKey)
        self.assertIsInstance(TestECCrypto.m2crypto_key, M2CryptoSK)
        self.assertIsInstance(TestECCrypto.m2crypto_key, M2CryptoPK)

    def test_generate_nacl(self):
        """
        Check if libnacl backend keys can be generated correctly.
        """
        self.assertIsInstance(TestECCrypto.libnacl_key, Key)
        self.assertIsInstance(TestECCrypto.libnacl_key, PrivateKey)
        self.assertIsInstance(TestECCrypto.libnacl_key, PublicKey)
        self.assertIsInstance(TestECCrypto.libnacl_key, LibNaCLSK)
        self.assertIsInstance(TestECCrypto.libnacl_key, LibNaCLPK)

    def test_generate_bogus(self):
        """
        Check if a bogus curve produces a RuntimeError
        """
        self.assertRaises(RuntimeError, self.ecc.generate_key, u"idontexist")

    def test_key_to_bin_m2crypto(self):
        """
        Check if ECCrypto correctly detects an M2Crypto key for bin.
        """
        key_bin = self.ecc.key_to_bin(TestECCrypto.m2crypto_key)

        self.assertEqual(key_bin, TestECCrypto.m2crypto_key.key_to_bin())

    def test_key_to_bin_nacl(self):
        """
        Check if ECCrypto correctly detects an libnacl key for bin.
        """
        key_bin = self.ecc.key_to_bin(TestECCrypto.libnacl_key)

        self.assertEqual(key_bin, TestECCrypto.libnacl_key.key_to_bin())

    def test_key_to_hash_m2crypto(self):
        """
        Check if ECCrypto correctly detects an M2Crypto key for hash.
        """
        key_hash = self.ecc.key_to_hash(TestECCrypto.m2crypto_key)

        self.assertEqual(key_hash, TestECCrypto.m2crypto_key.key_to_hash())

    def test_key_to_hash_nacl(self):
        """
        Check if ECCrypto correctly detects an libnacl key for hash.
        """
        key_hash = self.ecc.key_to_hash(TestECCrypto.libnacl_key)

        self.assertEqual(key_hash, TestECCrypto.libnacl_key.key_to_hash())

    def test_is_valid_private_bin_m2crypto(self):
        """
        Check if ECCrypto can detect a valid M2Crypto private key.
        """
        self.assertTrue(self.ecc.is_valid_private_bin(TestECCrypto.m2crypto_key.key_to_bin()))

    def test_is_valid_private_bin_m2crypto_public(self):
        """
        Check if ECCrypto doesn't detect a valid public M2Crypto key as a private key.
        """
        self.assertFalse(self.ecc.is_valid_private_bin(TestECCrypto.m2crypto_key.pub().key_to_bin()))

    def test_is_valid_private_bin_nacl(self):
        """
        Check if ECCrypto can detect a valid libnacl private key.
        """
        self.assertTrue(self.ecc.is_valid_private_bin(TestECCrypto.libnacl_key.key_to_bin()))

    def test_is_valid_private_bin_nacl_public(self):
        """
        Check if ECCrypto doesn't detect a valid public libnacl key as a private key.
        """
        self.assertFalse(self.ecc.is_valid_private_bin(TestECCrypto.libnacl_key.pub().key_to_bin()))

    def test_is_valid_public_bin_m2crypto(self):
        """
        Check if ECCrypto doesn't detect a valid M2Crypto private key as a public key.
        """
        self.assertFalse(self.ecc.is_valid_public_bin(TestECCrypto.m2crypto_key.key_to_bin()))

    def test_is_valid_public_bin_m2crypto_public(self):
        """
        Check if ECCrypto detects a valid public M2Crypto key as a public key.
        """
        self.assertTrue(self.ecc.is_valid_public_bin(TestECCrypto.m2crypto_key.pub().key_to_bin()))

    def test_is_valid_public_bin_nacl(self):
        """
        Check if ECCrypto doesn't detect a valid libnacl private key as a public key.
        """
        self.assertFalse(self.ecc.is_valid_public_bin(TestECCrypto.libnacl_key.key_to_bin()))

    def test_is_valid_public_bin_nacl_public(self):
        """
        Check if ECCrypto detects a valid public libnacl key as a public key.
        """
        self.assertTrue(self.ecc.is_valid_public_bin(TestECCrypto.libnacl_key.pub().key_to_bin()))



from __future__ import absolute_import

import os
from binascii import unhexlify

from twisted.internet.defer import inlineCallbacks

from ...base import MockIPv8, TestBase
from ....attestation.wallet.bonehexact.structs import BonehAttestation
from ....attestation.wallet.community import AttestationCommunity
from ....attestation.wallet.database import AttestationsDB
from ....attestation.wallet.pengbaorange.structs import PengBaoAttestation
from ....attestation.wallet.primitives.structs import BonehPrivateKey


class TestCommunity(TestBase):

    private_key = BonehPrivateKey.unserialize(unhexlify("01064c65dcb113f901064228da3ea57101064793a4f9c77901062b083e"
                                                        "8690fb0106408293c67e9f010601d1a9d3744901030f4243"))

    def setUp(self):
        super(TestCommunity, self).setUp()
        self.initialize(AttestationCommunity, 2)

    def create_node(self):
        return MockIPv8(u"curve25519", AttestationCommunity, working_directory=u":memory:")

    @inlineCallbacks
    def test_request_attestation_callback(self):
        """
        Check if the request_attestation callback is correctly called.
        """
        def f(peer, attribute_name, metadata):
            self.assertEqual(peer.address, self.nodes[1].endpoint.wan_address)
            self.assertEqual(attribute_name, "MyAttribute")
            self.assertDictEqual(metadata, {})

            f.called = True
        f.called = False

        yield self.introduce_nodes()

        self.nodes[0].overlay.set_attestation_request_callback(f)

        self.nodes[1].overlay.request_attestation(self.nodes[0].overlay.my_peer,
                                                  "MyAttribute",
                                                  TestCommunity.private_key)

        yield self.deliver_messages()

        self.assertTrue(f.called)
        # Request for attribute attestation goes unanswered
        self.nodes[1].overlay.request_cache.clear()

    @inlineCallbacks
    def test_request_attestation_callback_metadata(self):
        """
        Check if the request_attestation callback is correctly called with metadata.
        """

        def f(peer, attribute_name, metadata):
            self.assertEqual(peer.address, self.nodes[1].endpoint.wan_address)
            self.assertEqual(attribute_name, "MyAttribute")
            self.assertDictEqual(metadata, {u'test': u'123'})

            f.called = True

        f.called = False

        yield self.introduce_nodes()

        self.nodes[0].overlay.set_attestation_request_callback(f)

        self.nodes[1].overlay.request_attestation(self.nodes[0].overlay.my_peer,
                                                  "MyAttribute",
                                                  TestCommunity.private_key,
                                                  {'test': 123})

        yield self.deliver_messages()

        self.assertTrue(f.called)
        # Request for attribute attestation goes unanswered
        self.nodes[1].overlay.request_cache.clear()

    @inlineCallbacks
    def test_request_attestation(self):
        """
        Check if the request_attestation callback is correctly called.
        """
        def f(peer, attribute_name, _, __=None):
            self.assertEqual(peer.address, self.nodes[1].endpoint.wan_address)
            self.assertEqual(attribute_name, "MyAttribute")

            f.called = True
        f.called = False

        yield self.introduce_nodes()

        self.nodes[0].overlay.set_attestation_request_callback(lambda x, y, z: b"AttributeValue")
        self.nodes[0].overlay.set_attestation_request_complete_callback(f)

        self.nodes[1].overlay.request_attestation(self.nodes[0].overlay.my_peer,
                                                  "MyAttribute",
                                                  TestCommunity.private_key)

        yield self.deliver_messages(0.5)

        db_entries = self.nodes[1].overlay.database.get_all()
        self.assertEqual(1, len(db_entries))
        self.assertTrue(f.called)

    @inlineCallbacks
    def test_request_attestation_big(self):
        """
        Check if the request_attestation callback is correctly called for id_metadata_big.
        """

        def f(peer, attribute_name, _, __=None):
            self.assertEqual(peer.address, self.nodes[1].endpoint.wan_address)
            self.assertEqual(attribute_name, "MyAttribute")

            f.called = True

        f.called = False

        yield self.introduce_nodes()

        self.nodes[0].overlay.set_attestation_request_callback(lambda x, y, z: b"AttributeValue")
        self.nodes[0].overlay.set_attestation_request_complete_callback(f)

        self.nodes[1].overlay.request_attestation(self.nodes[0].overlay.my_peer,
                                                  "MyAttribute",
                                                  TestCommunity.private_key,
                                                  metadata={"id_format": "id_metadata_big"})

        yield self.deliver_messages(1.5)

        db_entries = self.nodes[1].overlay.database.get_all()
        self.assertEqual(1, len(db_entries))
        self.assertTrue(f.called)

    @inlineCallbacks
    def test_request_attestation_range(self):
        """
        Check if the request_attestation callback is correctly called for id_metadata_range_18plus.
        """

        def f(peer, attribute_name, _, __=None):
            self.assertEqual(peer.address, self.nodes[1].endpoint.wan_address)
            self.assertEqual(attribute_name, "MyAttribute")

            f.called = True

        f.called = False

        yield self.introduce_nodes()

        self.nodes[0].overlay.set_attestation_request_callback(lambda x, y, z: b"\x13")
        self.nodes[0].overlay.set_attestation_request_complete_callback(f)

        self.nodes[1].overlay.request_attestation(self.nodes[0].overlay.my_peer,
                                                  "MyAttribute",
                                                  TestCommunity.private_key,
                                                  metadata={"id_format": "id_metadata_range_18plus"})

        yield self.deliver_messages(2.0)

        db_entries = self.nodes[1].overlay.database.get_all()
        self.assertEqual(1, len(db_entries))
        self.assertTrue(f.called)

    @inlineCallbacks
    def test_verify_attestation(self):
        """
        Check if an attestation can be verified.
        """
        serialized = ""
        filename = os.path.join(os.path.dirname(__file__), 'attestation.txt')
        with open(filename, 'r') as f:
            serialized = unhexlify(f.read().strip())
        attestation = BonehAttestation.unserialize(serialized, "id_metadata")
        attestation_hash = unhexlify('9019195eb75c07ec3e86a62c314dcf5ef2bbcc0d')
        self.nodes[0].overlay.database.insert_attestation(attestation, attestation_hash, TestCommunity.private_key,
                                                          "id_metadata")
        self.nodes[0].overlay.attestation_keys[attestation_hash] = (TestCommunity.private_key, "id_metadata")

        def callback(rhash, values):
            self.assertEqual(attestation_hash, rhash)
            self.assertEqual(1, len(values))
            self.assertLess(0.99, values[0])
            callback.called = True
        callback.called = False
        self.nodes[1].overlay.verify_attestation_values(self.nodes[0].endpoint.wan_address,
                                                        attestation_hash,
                                                        [b"MyAttribute"],
                                                        callback,
                                                        "id_metadata")

        yield self.deliver_messages(0.5)

        self.assertTrue(callback.called)
        self.nodes[1].overlay.request_cache.clear()

    @inlineCallbacks
    def test_verify_attestation_big(self):
        """
        Check if an attestation can be verified for id_metadata_big.
        """
        filename = os.path.join(os.path.dirname(__file__), 'attestation_big.txt')
        with open(filename, 'r') as f:
            serialized = unhexlify(f.read().strip())
        attestation = BonehAttestation.unserialize(serialized, "id_metadata_big")
        attestation_hash = unhexlify('113d31c31b626268a16c198cbd58dd5aa8d1d81c')
        self.nodes[0].overlay.database.insert_attestation(attestation, attestation_hash, TestCommunity.private_key,
                                                          "id_metadata_big")
        self.nodes[0].overlay.attestation_keys[attestation_hash] = (TestCommunity.private_key, "id_metadata_big")

        def callback(rhash, values):
            self.assertEqual(attestation_hash, rhash)
            self.assertEqual(1, len(values))
            self.assertLess(0.99, values[0])
            callback.called = True

        callback.called = False
        self.nodes[1].overlay.verify_attestation_values(self.nodes[0].endpoint.wan_address,
                                                        attestation_hash,
                                                        [b"AttributeValue"],
                                                        callback,
                                                        "id_metadata_big")

        yield self.deliver_messages(1.5)

        self.assertTrue(callback.called)
        self.nodes[1].overlay.request_cache.clear()

    @inlineCallbacks
    def test_verify_attestation_range(self):
        """
        Check if an attestation can be verified for id_metadata_range_18plus.
        """
        filename = os.path.join(os.path.dirname(__file__), 'attestation_range.txt')
        with open(filename, 'r') as f:
            serialized = unhexlify(f.read().strip())
        attestation = PengBaoAttestation.unserialize_private(self.private_key, serialized, "id_metadata_range_18plus")
        attestation_hash = unhexlify('b40c8734ba6c91a49670c1f0152c7f4dac2a8272')
        self.nodes[0].overlay.database.insert_attestation(attestation, attestation_hash, TestCommunity.private_key,
                                                          "id_metadata_range_18plus")
        self.nodes[0].overlay.attestation_keys[attestation_hash] = (TestCommunity.private_key,
                                                                    "id_metadata_range_18plus")

        def callback(rhash, values):
            self.assertEqual(attestation_hash, rhash)
            self.assertEqual(1, len(values))
            self.assertLess(0.99, values[0])
            callback.called = True

        callback.called = False
        self.nodes[1].overlay.verify_attestation_values(self.nodes[0].endpoint.wan_address,
                                                        attestation_hash,
                                                        [b"\x01"],
                                                        callback,
                                                        "id_metadata_range_18plus")

        yield self.deliver_messages(2.5)

        self.assertTrue(callback.called)
        self.nodes[1].overlay.request_cache.clear()

    def test_load_key(self):
        """
        Check if we can load the community correctly after shut down.
        """
        # Write to a temporary folder.
        overlay = self.nodes[0].overlay
        temp_folder = self.temporary_directory()
        overlay.database = AttestationsDB(temp_folder, "test")

        # Create an attestation and write it to file.
        # Then close the database.
        attestation = BonehAttestation(TestCommunity.private_key.public_key(), [], "id_metadata")
        overlay.on_attestation_complete(attestation, TestCommunity.private_key, None, "test", b"a" * 20, "id_metadata")
        overlay.database.close(True)

        # Reload the community with the same database.
        self.nodes[0].overlay.__init__(self.nodes[0].my_peer, self.nodes[0].endpoint, self.nodes[0].network,
                                       working_directory=temp_folder, db_name="test")

        # The attestation should persist
        db_entries = self.nodes[0].overlay.database.get_all()
        self.assertEqual(1, len(db_entries))



from __future__ import absolute_import

import time

from twisted.internet.defer import inlineCallbacks, succeed

from ..base import TestBase
from ..mocking.ipv8 import MockIPv8
from ...dht.discovery import DHTDiscoveryCommunity
from ...dht.routing import Node


class TestDHTDiscoveryCommunity(TestBase):

    def setUp(self):
        super(TestDHTDiscoveryCommunity, self).setUp()
        self.initialize(DHTDiscoveryCommunity, 2)
        self.pinged = None
        self.puncture_to = None

        now = time.time()
        for node1 in self.nodes:
            node1.overlay.cancel_pending_task('store_peer')
            for node2 in self.nodes:
                if node1 == node2:
                    continue
                dht_node1 = Node(node1.my_peer.key, node1.my_peer.address)
                dht_node2 = Node(node2.my_peer.key, node2.my_peer.address)
                node1.overlay.tokens[dht_node2] = (now, node2.overlay.generate_token(dht_node1))

    def create_node(self, *args, **kwargs):
        return MockIPv8(u"curve25519", DHTDiscoveryCommunity)

    @inlineCallbacks
    def test_store_peer(self):
        yield self.introduce_nodes()
        yield self.nodes[0].overlay.store_peer()
        self.assertIn(self.nodes[0].my_peer.mid, self.nodes[1].overlay.store)
        self.assertIn(self.nodes[0].my_peer.mid, self.nodes[0].overlay.store_for_me)

    @inlineCallbacks
    def test_store_peer_fail(self):
        yield self.introduce_nodes()
        self.nodes[1].unload()
        d = self.nodes[0].overlay.store_peer()
        yield self.deliver_messages()
        self.assertFailure(d, RuntimeError)

    @inlineCallbacks
    def test_connect_peer(self):
        # Add a third node
        node = MockIPv8(u"curve25519", DHTDiscoveryCommunity)
        self.add_node_to_experiment(node)
        yield self.introduce_nodes()

        # Node1 is storing the peer of node0
        self.nodes[1].overlay.store[self.nodes[0].my_peer.mid].append(self.nodes[0].my_peer)
        self.nodes[0].overlay.store_for_me[self.nodes[0].my_peer.mid].append(self.nodes[1].my_peer)

        org_func = self.nodes[1].overlay.create_puncture_request

        def create_puncture_request(*args):
            self.puncture_to = args[1]
            return org_func(*args)
        self.nodes[1].overlay.create_puncture_request = create_puncture_request

        yield self.deliver_messages()
        nodes = yield self.nodes[2].overlay.connect_peer(self.nodes[0].my_peer.mid)
        self.assertEqual(self.puncture_to, self.nodes[2].my_peer.address)
        self.assertIn(self.nodes[0].overlay.my_peer.public_key.key_to_bin(),
                      [n.public_key.key_to_bin() for n in nodes])

    @inlineCallbacks
    def test_connect_peer_fail(self):
        yield self.introduce_nodes()
        self.nodes[1].unload()
        d = self.nodes[0].overlay.connect_peer(self.nodes[1].my_peer.mid)
        yield self.deliver_messages()
        self.assertFailure(d, RuntimeError)

    @inlineCallbacks
    def test_ping_pong(self):
        now = time.time() - 1

        node0 = Node(self.nodes[0].my_peer.key, self.nodes[0].my_peer.address)
        node0.last_response = now
        node0.last_queries.append(now)

        node1 = Node(self.nodes[1].my_peer.key, self.nodes[1].my_peer.address)
        node1.last_response = now
        node1.last_queries.append(now)

        key = node1.mid
        self.nodes[0].overlay.store[key].append(node1)
        self.nodes[1].overlay.store_for_me[key].append(node0)

        yield self.nodes[1].overlay.ping(node0)
        self.assertNotEqual(node0.last_response, now)
        self.assertNotEqual(node1.last_query, now)

    def test_ping_all(self):
        self.nodes[0].overlay.ping = lambda n: setattr(self, 'pinged', n) or succeed(None)

        node1 = Node(self.nodes[1].my_peer.key, self.nodes[1].my_peer.address)
        node1.last_response = time.time()
        node1.last_queries.append(time.time())

        self.nodes[0].overlay.store[node1.mid].append(node1)
        self.nodes[0].overlay.ping_all()
        self.assertIn(node1, self.nodes[0].overlay.store[node1.mid])

        node1.last_queries[-1] -= 100
        self.nodes[0].overlay.ping_all()
        self.assertNotIn(node1, self.nodes[0].overlay.store[node1.mid])

        self.nodes[0].overlay.store_for_me[node1.mid].append(node1)
        self.nodes[0].overlay.ping_all()
        self.assertEqual(self.pinged, None)
        self.assertIn(node1.mid, self.nodes[0].overlay.store_for_me)

        node1.last_response -= 30
        self.nodes[0].overlay.ping_all()
        self.assertEqual(self.pinged, node1)
        self.assertIn(node1, self.nodes[0].overlay.store_for_me[node1.mid])

        self.pinged = None
        node1.failed = 3
        self.nodes[0].overlay.ping_all()
        self.assertEqual(self.pinged, None)
        self.assertNotIn(node1, self.nodes[0].overlay.store_for_me[node1.mid])



from __future__ import absolute_import

from twisted.trial import unittest

from .....attestation.wallet.bonehexact.structs import BonehAttestation, BitPairAttestation
from .....attestation.wallet.primitives.structs import BonehPrivateKey, BonehPublicKey, pack_pair, unpack_pair
from .....attestation.wallet.primitives.value import FP2Value


class TestStructs(unittest.TestCase):

    def test_pack_pair(self):
        """
        Check if a pair of integers can be packed and unpacked correctly.
        """
        a = 21909485029845092485
        b = 9809809809800134013490

        packed = pack_pair(a, b)
        unpacked_a, unpacked_b, remainder = unpack_pair(packed)

        self.assertEqual(a, unpacked_a)
        self.assertEqual(b, unpacked_b)
        self.assertEqual(b'', remainder)

    def test_pack_pair_with_remainder(self):
        """
        Check if a pair of integers can be packed and unpacked correctly with a remainder.
        """
        a = 4684868464648654646385464634636
        b = 3212316546461687987934688
        remainder = b'iljaerlijerwfjilrwuj'

        packed = pack_pair(a, b) + remainder
        unpacked_a, unpacked_b, unpacked_remainder = unpack_pair(packed)

        self.assertEqual(a, unpacked_a)
        self.assertEqual(b, unpacked_b)
        self.assertEqual(remainder, unpacked_remainder)

    def test_serialize_boneh_pk(self):
        """
        Check if a BonehPublicKey can be serialized and unserialized correctly.
        """
        p = 884546864135123153155516635631631
        g = FP2Value(p, 58468546584635416356, 51468468484864846451)
        h = FP2Value(p, 651465444864846456151, 31213216564)
        key = BonehPublicKey(p, g, h)

        serialized = key.serialize()
        unserialized = BonehPublicKey.unserialize(serialized)

        self.assertEqual(p, unserialized.p)
        self.assertEqual(g, unserialized.g)
        self.assertEqual(h, unserialized.h)

    def test_serialize_boneh_pk_garbage(self):
        """
        Check if a BonehPublicKey returns None when unserializing garbage data.
        """
        unserialized = BonehPublicKey.unserialize(b'wowiuuorfuuj')

        self.assertEqual(None, unserialized)

    def test_serialize_boneh_pk_with_remainder(self):
        """
        Check if a BonehPublicKey can be serialized and unserialized correctly with remainder.
        """
        p = 69484635115151351513652987894784654156545665
        g = FP2Value(p, 68461115156531651631653163563, 132319884561841)
        h = FP2Value(p, 98781236511, 32185666658636546663635165635)
        key = BonehPublicKey(p, g, h)

        serialized = key.serialize() + b'iejriq94u2305ijiqnfa'
        unserialized = BonehPublicKey.unserialize(serialized)

        self.assertEqual(p, unserialized.p)
        self.assertEqual(g, unserialized.g)
        self.assertEqual(h, unserialized.h)

    def test_serialize_boneh_sk(self):
        """
        Check if a BonehPrivateKey can be serialized and unserialized correctly.
        """
        n = 684684664663646413641656355663
        p = 13121354644565465
        g = FP2Value(p, 96874984765651564, 3216465486956184847984564)
        h = FP2Value(p, 987987984, 8794181684561484515646518456148156454544)
        t1 = 6848489484665156651566515665163565636663563
        key = BonehPrivateKey(p, g, h, n, t1)

        serialized = key.serialize()
        unserialized = BonehPrivateKey.unserialize(serialized)

        self.assertEqual(n, unserialized.n)
        self.assertEqual(p, unserialized.p)
        self.assertEqual(g, unserialized.g)
        self.assertEqual(h, unserialized.h)
        self.assertEqual(t1, unserialized.t1)

    def test_serialize_boneh_sk_with_remainder(self):
        """
        Check if a BonehPrivateKey can be serialized and unserialized correctly with remainder.
        """
        n = 8441516511511
        p = 6415165156531635115353115155153515335333521
        g = FP2Value(p, 86416515655841555261455, 84986446854656165)
        h = FP2Value(p, 156151352151313613651, 84686815661132515513313)
        t1 = 321321564645164653164687
        key = BonehPrivateKey(p, g, h, n, t1)

        serialized = key.serialize() + b';k;mezrako;erjiragtijtgrioj'
        unserialized = BonehPrivateKey.unserialize(serialized)

        self.assertEqual(n, unserialized.n)
        self.assertEqual(p, unserialized.p)
        self.assertEqual(g, unserialized.g)
        self.assertEqual(h, unserialized.h)
        self.assertEqual(t1, unserialized.t1)

    def test_serialize_boneh_sk_garbage(self):
        """
        Check if a BonehPrivateKey returns None when unserializing garbage data.
        """
        unserialized = BonehPrivateKey.unserialize(b'wowiuuorfuuj')

        self.assertEqual(None, unserialized)

    def test_serialize_bitpair_attestation(self):
        """
        Check if a BitPairAttestation can be serialized and unserialized correctly.
        """
        p = 64151651565316351153531151551535153353335217
        a = FP2Value(p, 86416515655841555261455, 84986446854656165)
        b = FP2Value(p, 156151352151313613651, 84686815661132515513313)
        c = FP2Value(p, 8468548546854654, 6846843513131351)
        attest = BitPairAttestation(a, b, c)

        serialized = attest.serialize()
        unserialized = BitPairAttestation.unserialize(serialized, p)

        self.assertEqual(a, unserialized.a)
        self.assertEqual(b, unserialized.b)
        self.assertEqual(c, unserialized.complement)

    def test_serialize_bitpair_attestation_with_remainder(self):
        """
        Check if a BitPairAttestation can be serialized and unserialized correctly with remainder.
        """
        p = 6416514551355336516355565655656541131
        a = FP2Value(p, 64168498535685461646, 6846513216165165)
        b = FP2Value(p, 32184987216545974, 987984116541354132132165464)
        c = FP2Value(p, 13265, 848464848609849840645102)
        attest = BitPairAttestation(a, b, c)

        serialized = attest.serialize() + b'ih43qo8hepi9fjaegihngatejlidgvujhnmk '
        unserialized = BitPairAttestation.unserialize(serialized, p)

        self.assertEqual(a, unserialized.a)
        self.assertEqual(b, unserialized.b)
        self.assertEqual(c, unserialized.complement)

    def test_serialize_bitpair_compress(self):
        """
        Check if a BitPairAttestation can be compressed for proofing.
        """
        p = 64151651565316351153531151551535153353335217
        a = FP2Value(p, 86416515655841555261455, 84986446854656165)
        b = FP2Value(p, 156151352151313613651, 84686815661132515513313)
        c = FP2Value(p, 8468548546854654, 6846843513131351)
        attest = BitPairAttestation(a, b, c)

        self.assertEqual(a * b * c, attest.compress())

    def test_serialize_attestation_empty(self):
        """
        Check if an Attestation can be serialized and unserialized correctly with no bitpairs.
        """
        p = 884546864135123153155516635631631
        g = FP2Value(p, 58468546584635416356, 51468468484864846451)
        h = FP2Value(p, 651465444864846456151, 31213216564)
        key = BonehPublicKey(p, g, h)
        bitpairs = []
        attest = BonehAttestation(key, bitpairs)

        serialized = attest.serialize()
        unserialized = BonehAttestation.unserialize(serialized)

        self.assertEqual(p, unserialized.PK.p)
        self.assertEqual(g, unserialized.PK.g)
        self.assertEqual(h, unserialized.PK.h)
        self.assertListEqual(bitpairs, unserialized.bitpairs)

    def test_serialize_attestation_one(self):
        """
        Check if an Attestation can be serialized and unserialized correctly with one bitpair.
        """
        p = 884546864135123153155516635631631
        g = FP2Value(p, 58468546584635416356, 51468468484864846451)
        h = FP2Value(p, 651465444864846456151, 31213216564)
        a = FP2Value(p, 64168498535685461646, 6846513216165165)
        b = FP2Value(p, 32184987216545974, 987984116541354132132165464)
        c = FP2Value(p, 13265, 848464848609849840645102)
        bitpair = BitPairAttestation(a, b, c)
        key = BonehPublicKey(p, g, h)
        bitpairs = [bitpair]
        attest = BonehAttestation(key, bitpairs)

        serialized = attest.serialize()
        unserialized = BonehAttestation.unserialize(serialized)

        self.assertEqual(p, unserialized.PK.p)
        self.assertEqual(g, unserialized.PK.g)
        self.assertEqual(h, unserialized.PK.h)
        for bp in unserialized.bitpairs:
            self.assertEqual(bitpair.a, bp.a)
            self.assertEqual(bitpair.b, bp.b)
            self.assertEqual(bitpair.complement, bp.complement)

    def test_serialize_attestation_many(self):
        """
        Check if an Attestation can be serialized and unserialized correctly with twenty bitpairs.
        """
        p = 884546864135123153155516635631631
        g = FP2Value(p, 58468546584635416356, 51468468484864846451)
        h = FP2Value(p, 651465444864846456151, 31213216564)
        a = FP2Value(p, 64168498535685461646, 6846513216165165)
        b = FP2Value(p, 32184987216545974, 987984116541354132132165464)
        c = FP2Value(p, 13265, 848464848609849840645102)
        bitpair = BitPairAttestation(a, b, c)
        key = BonehPublicKey(p, g, h)
        bitpairs = [bitpair] * 20
        attest = BonehAttestation(key, bitpairs)

        serialized = attest.serialize()
        unserialized = BonehAttestation.unserialize(serialized)

        self.assertEqual(p, unserialized.PK.p)
        self.assertEqual(g, unserialized.PK.g)
        self.assertEqual(h, unserialized.PK.h)
        for bp in unserialized.bitpairs:
            self.assertEqual(bitpair.a, bp.a)
            self.assertEqual(bitpair.b, bp.b)
            self.assertEqual(bitpair.complement, bp.complement)



from __future__ import absolute_import

import logging
from traceback import format_tb

from twisted.internet import reactor
from twisted.internet.defer import maybeDeferred
from twisted.python.compat import intToBytes
from twisted.web import http, server

from .json_util import dumps
from .root_endpoint import RootEndpoint
from ..taskmanager import TaskManager


class RESTManager(TaskManager):
    """
    This class is responsible for managing the startup and closing of the HTTP API.
    """

    def __init__(self, session):
        super(RESTManager, self).__init__()
        self._logger = logging.getLogger(self.__class__.__name__)
        self.session = session
        self.site = None
        self.root_endpoint = None

    def start(self, port=8085):
        """
        Starts the HTTP API with the listen port as specified in the session configuration.
        """
        self.root_endpoint = RootEndpoint(self.session)
        site = server.Site(resource=self.root_endpoint)
        site.requestFactory = RESTRequest
        self.site = reactor.listenTCP(port, site, interface="127.0.0.1")

    def stop(self):
        """
        Stop the HTTP API and return a deferred that fires when the server has shut down.
        """
        return maybeDeferred(self.site.stopListening)


class RESTRequest(server.Request):
    """
    This class gracefully takes care of unhandled exceptions raised during the processing of any request.
    """
    defaultContentType = b"application/json"

    def __init__(self, *args, **kw):
        server.Request.__init__(self, *args, **kw)
        self._logger = logging.getLogger(self.__class__.__name__)

        # Default headers
        self.setHeader(b'Access-Control-Allow-Origin', b'*')

    def processingFailed(self, failure):
        self._logger.exception(failure)
        response = {
            u"error": {
                u"handled": False,
                u"code": failure.value.__class__.__name__,
                u"message": str(failure.value)
            }
        }
        if self.site.displayTracebacks:
            response[u"error"][u"trace"] = format_tb(failure.getTracebackObject())

        body = dumps(response, True).encode('utf-8')
        self.setResponseCode(http.INTERNAL_SERVER_ERROR)
        self.setHeader(b'Content-Type', self.defaultContentType)
        self.setHeader(b'Content-Length', intToBytes(len(body)))
        self.write(body)
        self.finish()
        return failure



from __future__ import absolute_import

import abc
from hashlib import sha1
import six


class Key(six.with_metaclass(abc.ABCMeta, object)):
    """
    Interface for a public or private key.
    """

    @abc.abstractmethod
    def pub(self):
        pass

    @abc.abstractmethod
    def has_secret_key(self):
        pass

    @abc.abstractmethod
    def key_to_bin(self):
        pass

    def key_to_hash(self):
        if self.has_secret_key():
            return sha1(self.pub().key_to_bin()).digest()
        return sha1(self.key_to_bin()).digest()


class PrivateKey(six.with_metaclass(abc.ABCMeta, Key)):
    """
    Interface for a private key.
    """

    def has_secret_key(self):
        return True

    @abc.abstractmethod
    def signature(self, msg):
        pass


class PublicKey(six.with_metaclass(abc.ABCMeta, Key)):
    """
    Interface for a public key.
    """

    def pub(self):
        return self

    def has_secret_key(self):
        return False

    @abc.abstractmethod
    def verify(self, signature, msg):
        pass

    @abc.abstractmethod
    def get_signature_length(self):
        pass
